import gzip
import json
import os
import shutil


def copy_file(src_file, dst_file):
    """
    Copy a single file from source to destination

    Parameters:
    - src_file (str): The path to the source file
    - dst_file (str): The path to the destination file
    """
    dst_dir = os.path.dirname(dst_file)
    os.makedirs(dst_dir, exist_ok=True)
    shutil.copyfile(src_file, dst_file)


def copy_external_structures(src_json: str, destination: str) -> None:
    """
    Copy external structure JSON files referenced by ``src_json``
    into the same ``_datasets`` tree as ``destination``, preserving relative paths.
    """

    try:
        if src_json.endswith(".gz"):
            fd = gzip.open(src_json, "rt", encoding="utf8")
        else:
            fd = open(src_json, "r", encoding="utf8")
        dataset = json.load(fd)
        fd.close()
    except Exception:
        # if anything goes wrong reading the dataset, just skip
        print("Warning: could not read dataset JSON to copy external structures")
        return

    if not isinstance(dataset, dict):
        return

    structures = dataset.get("structures", [])
    if not isinstance(structures, list):
        return

    # return if there are no external structures
    if "data" not in structures[0]:
        return

    src_base = os.path.dirname(src_json)

    for struct in structures:
        if not isinstance(struct, dict):
            continue

        data = struct["data"]

        # Expecting data to be just the path of the structure.json
        if isinstance(data, str):
            rel_path = data
        else:
            raise ValueError(
                "unsupported external structure data format: "
                f"expected string but got {data.__class__.__name__}"
            )

        rel_path = os.path.normpath(rel_path)

        # Source file: relative to *original* JSON location
        src_struct = os.path.join(src_base, rel_path)

        if not os.path.exists(src_struct):
            raise FileNotFoundError(f"External structure file {src_struct} missing")

        # Destination: under the target destination folder
        dst_struct = os.path.join(destination, rel_path)
        copy_file(src_struct, dst_struct)



from typing import Dict, List, Tuple, Union


try:
    from stk import Molecule

    HAVE_STK = True
except ImportError:

    class Molecule:
        pass

    HAVE_STK = False


def _stk_valid_structures(
    structures: Union[Molecule, List[Molecule]],
) -> Tuple[List[Molecule], bool]:
    if HAVE_STK and isinstance(structures, Molecule):
        # deal with the user passing a single structure
        return [structures], True
    elif (
        HAVE_STK
        and isinstance(structures, list)
        and isinstance(structures[0], Molecule)
    ):
        for structure in structures:
            assert isinstance(structure, Molecule)
        return structures, True
    else:
        return structures, False


def _stk_to_json(molecule: Molecule) -> Dict[str, Union[int, list]]:
    """Implementation of structures_to_json for stk's ``Molecule``.

    The current implementation assumes no periodic information, which is safe
    for the majority of stk molecules. If necessary, we can add cell information
    in the future.

    """

    # stk specific bond order map.
    bond_map = {
        # stk int to chemiscope int.
        1: 1,
        2: 2,
        3: 3,
        # Represents dative bond types in stk.
        9: 1,
    }

    pos_mat = molecule.get_position_matrix()
    data = {}
    data["size"] = molecule.get_num_atoms()
    data["names"] = [atom.__class__.__name__ for atom in molecule.get_atoms()]
    data["x"] = [float(pos_mat[atom.get_id()][0]) for atom in molecule.get_atoms()]
    data["y"] = [float(pos_mat[atom.get_id()][1]) for atom in molecule.get_atoms()]
    data["z"] = [float(pos_mat[atom.get_id()][2]) for atom in molecule.get_atoms()]
    try:
        data["bonds"] = [
            (
                bond.get_atom1().get_id(),
                bond.get_atom2().get_id(),
                bond_map[bond.get_order()],
            )
            for bond in molecule.get_bonds()
        ]
    except KeyError as e:
        raise ValueError(f"Unknown bond order {e} (1, 2, 3, 9 (in stk) allowed)") from e

    return data


def _stk_all_atomic_environments(
    structures: List[Molecule],
    cutoff: float,
) -> List[Tuple[int, int, float]]:
    "Extract all atomic environments out of a set of stk Molecule objects"
    environments = []
    for structure_i, structure in enumerate(structures):
        for atom in structure.get_atoms():
            environments.append((structure_i, atom.get_id(), cutoff))

    return environments



import warnings

from ..jupyter import show
from ..structures import extract_properties
from ._metatomic import metatomic_featurizer


__all__ = ["explore", "metatomic_featurizer", "get_featurizer"]

KNOWN_FEATURIZERS = ["pet-mad-1.0"]


def get_featurizer(name):
    """
    Get a featurizer by name for feature extraction. Currently available version is:
    "pet-mad-1.0", which returns an instance of `PETMADFeaturizer
    <https://arxiv.org/abs/2506.19674>`_.

    :param str name: name of the featurizer. Must match one of the known versions.
        Currently available is "pet-mad-1.0"

    .. warning::

        This function requires additional dependencies. Install them using:

        .. code:: bash

            pip install chemiscope[explore]
    """
    try:
        import torch
        from pet_mad.explore import PETMADFeaturizer
    except ImportError as e:
        raise ImportError(
            "Required package not found. Please install the dependencies with "
            "`pip install chemiscope[explore]`."
        ) from e

    if not isinstance(name, str):
        raise TypeError(f"featurizer name must be a string, not {type(name)}")

    if name == "pet-mad-1.0":
        device = "cuda" if torch.cuda.is_available() else "cpu"
        model = PETMADFeaturizer(version="1.0.0", device=device, batch_size=1)
        return model
    else:
        raise ValueError(
            f"unknown featurizer: {name}. Available options "
            f"are: {', '.join(KNOWN_FEATURIZERS)}"
        )


def explore(
    structures=None,
    featurizer=None,
    properties=None,
    environments=None,
    settings=None,
    mode="default",
    write_input=None,
    *,
    frames=None,
):
    """
    Automatically generate an interactive Chemiscope visualization of atomic structures.

    This function creates a low-dimensional representation of the input ``structures``
    and displays them using a Chemiscope widget. It supports automatic featurization
    with `PETMADFeaturizer <https://arxiv.org/abs/2506.19674>`_ or a custom
    featurization function.

    The default ``PETMADFeaturizer`` computes `PET-MAD
    <https://arxiv.org/abs/2503.14118>`_ features from the structures and projects them
    into the 3D MAD latent space.

    If available, all properties are extracted automatically from the structures.

    If one does not specify a ``featurizer`` (or sets it as a ``None``), only properties
    will be displayed on the map visualizer panel, as long as there are at least two of
    them.

    Overall, the visualization can include: properties extracted from the structures,
    additional user-provided properties, features from either the built-in PET-MAD
    featurizer, with dimensionality reduction, or a custom user-provided featurization
    functions.

    :param list structures: list of structures

    :param featurizer: either string specifying a featurizer version (currently only
        'pet-mad-1.0'), a custom callable function, or None. Used to compute features
        and perform dimensionality reduction on the ``structures``. For automatic
        default option, use ``pet-mad-1.0``. The callable should take ``structures`` as
        the first argument and ``environments`` as the second argument. The return value
        must be a features array of shape ``(n_structures, n_features)`` if
        ``environments`` is ``None``, or ``(n_environments, n_features)`` otherwise.

    :param dict properties: optional. Additional properties to be included in the
        visualization. This dictionary can contain any other relevant data associated
        with the atomic structures. Properties can be extracted from structures with
        :py:func:`extract_properties` or manually defined by the user.

    :param environments: optional. List of environments (described as ``(structure id,
        center id, cutoff)``) to include when extracting the atomic properties. Can be
        extracted from structures with :py:func:`all_atomic_environments` or manually
        defined.

    :param dict settings: optional dictionary of settings to use when displaying the
        data. Possible entries for the ``settings`` dictionary are documented in the
        chemiscope input file reference.

    :param str mode: optional. Visualization mode for the chemiscope widget. Can be one
        of "default", "structure", or "map". The default mode is "default".

    :param str device: torch device to use for the calculation with the default
            ``PETMADFeaturizer``. If `None`, we will try the options in the model's
            ``supported_device`` in order.

    :param int batch_size: optional. Number of structures processed in each batch with
        the default ``PETMADFeaturizer``.

    :param string write_input: optional. A path to save the chemiscope input file
        created by this function. Afterwards, the file can be loaded using
        :py:func:`chemiscope.show_input`

    :return: a chemiscope widget for interactive visualization

    To use this function, additional dependencies are required, specifically, `pet_mad`_

    :return: a chemiscope widget for interactive visualization

    To use this function, additional dependencies are required, specifically, `pet-mad`_
    used for the default dimensionality reduction. They can be installed with the
    following command:

    .. code:: bash

        pip install chemiscope[explore]

    Here is an example using this function with and without a featurizer function. The
    structures are obtained by reading the structures from a file that `ase <ase-io_>`_
    can read, and performing Kernel PCA using `sklearn`_ on a descriptor computed with
    SOAP using the `dscribe`_ library.

    .. code-block:: python

        import chemiscope
        import ase.io
        import dscribe.descriptors
        import sklearn.decomposition

        # Read the structures from the dataset
        structures = ase.io.read("trajectory.xyz", ":")

        # 1) Basic usage with default featurizer (PET-MAD featurization + Sketch-Map)
        chemiscope.explore(structures, featurizer="pet-mad-1.0")

        # or
        featurizer = chemiscope.get_featurizer("pet-mad-1.0")
        chemiscope.explore(structures, featurizer=featurizer)


        # Define a function for dimensionality reduction
        def soap_kpca_featurize(structures, environments):
            if environments is not None:
                raise ValueError("'environments' are not supported by this featurizer")
            # Compute descriptors
            soap = dscribe.descriptors.SOAP(
                species=["C"],
                r_cut=4.5,
                n_max=8,
                l_max=6,
                periodic=True,
            )
            descriptors = soap.create(structures)

            # Apply KPCA
            kpca = sklearn.decomposition.KernelPCA(n_components=2, gamma=0.05)

            # Return a 2D array of reduced features
            return kpca.fit_transform(descriptors)


        # 2) Example with a custom featurizer function
        chemiscope.explore(structures, featurizer=soap_kpca_featurize)

    For more examples, see the related `documentation <chemiscope-explore_>`_.

    .. _ase-io: https://wiki.fysik.dtu.dk/ase/ase/io/io.html
    .. _pet_mad: https://github.com/lab-cosmo/pet-mad
    .. _sklearn: https://scikit-learn.org/
    .. _dscribe: https://singroup.github.io/dscribe/latest/
    .. _chemiscope-explore: https://chemiscope.org/docs/examples/6-explore.html
    """

    if frames is not None:
        warnings.warn(
            "`frames` argument is deprecated, use `structures` instead",
            stacklevel=2,
        )
        if structures is not None:
            raise ValueError("cannot use both `structures` and `frames` arguments")

        structures = frames

    merged_properties = {}

    if isinstance(featurizer, str):
        featurizer_instance = get_featurizer(featurizer)
        merged_properties["features"] = featurizer_instance(structures, environments)
    elif callable(featurizer):
        merged_properties["features"] = featurizer(structures, environments)

    merged_properties.update(extract_properties(structures, environments))

    if properties is not None:
        merged_properties.update(properties)

    if mode != "structure":
        if _count_effective_properties(merged_properties) < 2:
            raise ValueError(
                "To display the widget, you need to provide at least two properties to "
                "plot: either pass explicit properties using the 'properties' argument,"
                " where each dimension counts as a separate property, or use a "
                "featurizer (e.g. 'pet-mad-1.0') that generates properties (features)"
            )

    widget = show(
        structures=structures,
        properties=merged_properties,
        shapes=None,
        environments=environments,
        settings=settings,
        mode=mode,
    )

    if write_input is not None:
        widget.save(write_input)

    return widget


def _count_effective_properties(properties):
    """
    Count the total number of properties that can we displayed in the widget,
    where each dimension counts as a separate property
    """
    count = 0
    for value in properties.values():
        if hasattr(value, "shape") and len(value.shape) > 1:
            # for multi-dimentional props, each will be displayed as a separate property
            # in the widget
            count += value.shape[1]
        else:
            count += 1
    return count



import uuid

from sphinx.errors import ExtensionError


class FilePathIterator:
    """
    Generates unique file names with a defined pattern.
    The file names consist of a prefix, an infix (UUID by default), a counter,
    and an extension.

    Used in the ChemiscopeScraper to create file names for the saved datasets
    next to the generated .rst files (docs/src/examples/_datasets/)
    """

    def __init__(self, prefix="fig", extension=".json.gz"):
        """
        Initialize the iterator

        Parameters:
        - prefix (str): The prefix for image filenames
        - extension (str): The file extension for files
        """
        self.prefix = prefix
        self.extension = extension
        self.counter = 0
        self._stop = 1000000
        self.infix = uuid.uuid4()

    def __iter__(self):
        """Iterate over file paths generated by the iterator"""
        for _i in range(self._stop):
            yield self.next()
        else:
            raise ExtensionError(f"Generated over {self._stop} files")

    def __next__(self):
        """Generates the file name"""
        self.counter += 1
        return f"{self.prefix}_{self.infix}_{self.counter:03d}{self.extension}"

    def next(self):
        """Return the next file path, with numbering starting at 1"""
        return self.__next__()

    def set_infix(self, infix):
        """
        Set a new infix identifier to be used in the generated filename

        Parameters:
        - infix (str): The string used as an infix
        """
        self.infix = infix



import os
import warnings

from chemiscope.jupyter import ChemiscopeWidget, MapWidget, StructureWidget

from ._file_path_iterator import FilePathIterator
from ._utils import copy_external_structures


warnings.filterwarnings(
    "ignore",
    message="chemiscope.show only works in a jupyter notebook or a sphinx build",
)


class ChemiscopeScraper:
    """Custom scraper for Chemiscope visualizations"""

    def __init__(self):
        # Create an iterator to generate the file name
        self.iterator = FilePathIterator()

        # Process the widgets once by storing their `id()` here
        self.seen = set()

    def __repr__(self):
        return "ChemiscopeScraper"

    def __call__(self, _block, block_vars, _gallery_conf):
        """
        Extracts Chemiscope widget data, saves its .json dataset in the .rst directory
        and generates a .rst directive for embedding. Also copies external structures
        if present.

        Triggered on the each output of the script.

        Parameters:
        - _block: Unused parameter, a .rst code block
        - block_vars (dict): Variables from the executed code block
        - _gallery_conf: Unused parameter, sphinx-gallery config

        Returns:
        - str: .rst directive for embedding the chemiscope visualization
        """
        # Retrieve the chemiscope widget from block variables
        widget = block_vars.get("example_globals", {}).get("___")
        mode = self.get_widget_mode(widget)

        if (id(widget)) not in self.seen and mode is not None:
            self.seen.add(id(widget))

            # Get the target directory to save the dataset next to the .rst files
            target_file = block_vars.get("target_file")  # Python file
            rst_dataset_dir = os.path.join(os.path.dirname(target_file), "_datasets")
            os.makedirs(rst_dataset_dir, exist_ok=True)

            # Save the related dataset to the directory next to the related .rst file
            infix = os.path.splitext(os.path.basename(target_file))[0]
            dataset_filename = self.get_dataset_filename(infix)
            dataset_file_path = os.path.join(rst_dataset_dir, dataset_filename)
            widget.save(dataset_file_path)

            # also saves in-place to facilitate copying external structures
            src_file = block_vars.get("src_file")
            src_dataset = os.path.join(os.path.dirname(src_file), dataset_filename)
            widget.save(src_dataset)

            copy_external_structures(src_dataset, rst_dataset_dir)

            rel_file_path = os.path.join("_datasets", dataset_filename)

            # Return the .rst directive
            return self.generate_rst(rel_file_path, mode)
        else:
            # chemiscope.show is not called
            return ""

    def generate_rst(self, filepath, mode):
        """Generate a .rst directive for embedding chemiscope widget"""
        return f""".. chemiscope:: {filepath}
            :mode: {mode}
            :warning_timeout: {2000}
        """

    def get_widget_mode(self, widget):
        """Determine the mode of the chemiscope widget"""
        if isinstance(widget, ChemiscopeWidget):
            return "default"
        elif isinstance(widget, StructureWidget):
            return "structure"
        elif isinstance(widget, MapWidget):
            return "map"
        else:
            return None

    def get_dataset_filename(self, infix):
        """Generate a dataset filename using the file path infix"""
        self.iterator.set_infix(infix)
        return self.iterator.next()



from . import streamlit  # noqa: F401
from .explore import explore, get_featurizer, metatomic_featurizer  # noqa: F401
from .input import (  # noqa: F401
    create_input,
    quick_settings,
    write_external_structures,
    write_input,
)
from .jupyter import show, show_input  # noqa
from .structures import (  # noqa: F401
    all_atomic_environments,
    arrow_from_vector,
    ase_tensors_to_ellipsoids,
    ase_vectors_to_arrows,
    ellipsoid_from_tensor,
    extract_properties,
)
from .version import __version__  # noqa: F401



# -*- coding: utf-8 -*-
import asyncio
import base64
import gzip
import itertools
import json
import warnings
from pathlib import Path

import ipywidgets
from traitlets import Bool, Dict, Int, Unicode

from .input import create_input
from .version import __version__


# this needs to match the version/name defined in
# python/jupyter/src/labextension.ts
PACKAGE_NAME = "chemiscope"
PACKAGE_VERSION = __version__


class ChemiscopeWidgetBase(ipywidgets.DOMWidget, ipywidgets.ValueWidget):
    _view_module = Unicode(PACKAGE_NAME).tag(sync=True)
    _view_module_version = Unicode(PACKAGE_VERSION).tag(sync=True)
    value = Unicode().tag(sync=True)
    has_metadata = Bool().tag(sync=True)

    # synchronized settings from the JS side. You can assign to this field to
    # change what's being displayed by chemiscope, but you need to assign a full
    # dictionary (`widget.settings["map"]["x"]["property"] = "foo"` will not
    # work, but `widget.settings = updated_settings` will).
    settings = Dict().tag(sync=True)
    # timeout for warning messages
    warning_timeout = Int(10000).tag(sync=True)
    # currently selected environment/structure/atom
    selected_ids = Dict().tag(sync=True)
    # index of the currently active viewer in the grid
    active_viewer = Int(0).tag(sync=True)

    def __init__(
        self, data, has_metadata, warning_timeout=10000, cache_structures=True
    ):
        super().__init__()
        self.has_metadata = has_metadata
        if "settings" in data:
            self.settings = data["settings"]

        # timeout for warning messages (ms). 0 to make persistent, -1 to disable
        self.warning_timeout = warning_timeout

        self._request_counter = itertools.count()
        self._pending_requests = {}

        # hold structures on the python side to save js memory
        self._structures_cache = {}
        if cache_structures:
            data["structures"] = self._cache_structures(data["structures"])

        # pass data to javascript through a traitlet
        self.value = json.dumps(data)

        # listen for custom messages from the JS side
        self.on_msg(self._handle_js_msg)

    def _cache_structures(self, structures):
        """Cache full structures and return structure references"""
        if not structures or "data" in structures[0]:
            return structures

        structure_refs = []
        for i, structure in enumerate(structures):
            if "data" in structure:
                raise ValueError("Got a mix of explicit and external structures")

            cache_key = f"structure-{i}"
            self._structures_cache[cache_key] = structure
            structure_refs.append({"size": structure["size"], "data": cache_key})

        return structure_refs

    def get_map_image(self):
        """
        Request a snapshot of the map. Returns a Future that resolves to the
        image data (PNG formatted).

        This method is asynchronous. In a Jupyter notebook, you should ``await``
        it to get the data.

        .. code-block:: python

            # Get raw image data
            data = await widget.get_map_image()

            # Display it
            from IPython.display import Image

            display(Image(data))

        :return: A Future that resolves to the image data as bytes.
        """

        if self._view_name == "StructureView":
            raise RuntimeError(
                "Cannot retrieve map image: this widget is a structure-only viewer."
            )

        return self._request_screenshot("map")

    def get_structure_image(self):
        """
        Request a snapshot of the active structure viewer.
        Returns a Future that resolves to the image data (PNG formatted).

        This method is asynchronous. In a Jupyter notebook,
        you should ``await`` it to get the data.

        .. code-block:: python

            data = await widget.get_structure_image()

        :return: A Future that resolves to the image data as bytes.
        """

        if self._view_name == "MapView":
            raise RuntimeError(
                "Cannot retrieve structure image: this widget is a map-only viewer."
            )

        return self._request_screenshot("structure")

    def save_map_image(self, path):
        """
        Save a snapshot of the map to a file.

        This method starts a background task to save the
        image. You can ``await`` it if you need to ensure the
        file is written before proceeding.

        .. code-block:: python

            widget.save_map_image("map.png")

        :param str path: Path where the image will be saved.
        :return: A Future that resolves when the file is written.
        """

        if self._view_name == "StructureView":
            raise RuntimeError(
                "Cannot retrieve map image: this widget is a structure-only viewer."
            )

        return asyncio.ensure_future(self._save_image_to_file(path, "map"))

    def save_structure_image(self, path):
        """
        Save a snapshot of the active structure viewer to a file.

        This method starts a background task to save the image.
        You can ``await`` it if you need to ensure the file is
        written before proceeding.

        .. code-block:: python

            widget.save_structure_image("structure.png")

        :param str path: Path where the image will be saved.
        :return: A Future that resolves when the file is written.
        """

        if self._view_name == "MapView":
            raise RuntimeError(
                "Cannot save structure image: this widget is a map-only viewer."
            )

        return asyncio.ensure_future(self._save_image_to_file(path, "structure"))

    def get_structure_sequence(self, indices, settings=None):
        """
        Request a sequence of structure snapshots. Returns a Future that
        resolves to a list of image data (PNG formatted).

        This allows rendering multiple frames efficiently without blocking
        the UI. The sequence is processed in the browser.

        The ``indices`` list can contain integers (structure index) or
        dictionaries specifying ``structure`` and ``atom`` indices
        (for environments).

        The ``settings`` list, if provided, must have the same length
        as ``indices``. Each element is a dictionary of structure
        settings (e.g. ``{"spaceFilling": True}``) to apply for
        that specific frame.

        .. code-block:: python

            indices = [0, 1, 2]
            settings = [{"spaceFilling": True}, {}, {"spaceFilling": False}]
            data_list = await widget.get_structure_sequence(indices, settings)

        :param list indices: List of indices (int or dict) to render.
        :param list settings: Optional list of settings dicts to apply
            for each frame.
        :return: A Future that resolves to a list of image data bytes.
        """

        if self._view_name == "MapView":
            raise RuntimeError(
                "Cannot save structure image: this widget is a map-only viewer."
            )

        return asyncio.ensure_future(
            self._process_structure_sequence(indices, settings)
        )

    def save_structure_sequence(self, indices, paths, settings=None):
        """
        Save a sequence of structure snapshots to files.

        This method acts as a wrapper around :meth:`get_structure_sequence`
        and writes the results to files.

        .. code-block:: python

            indices = [0, 10, 20]
            paths = ["frame_0.png", "frame_10.png", "frame_20.png"]
            await widget.save_structure_sequence(indices, paths)

        :param list indices: List of indices (int or dict) to render.
        :param list paths: List of file paths where images will be saved.
            Must match length of ``indices``.
        :param list settings: Optional list of settings dicts to apply
            to each frame.
        :return: A Future that resolves when all files are written.
        """
        if len(indices) != len(paths):
            raise ValueError("indices and paths must have the same length")
        if settings is not None and len(indices) != len(settings):
            raise ValueError("indices and settings must have the same length")
        if self._view_name == "MapView":
            raise RuntimeError(
                "Cannot save structure image: this widget is a map-only viewer."
            )

        async def _save_impl():
            data_list = await self.get_structure_sequence(indices, settings)
            for path, data in zip(paths, data_list, strict=True):
                with open(path, "wb") as f:
                    f.write(data)

        return asyncio.ensure_future(_save_impl())

    def _request_screenshot(self, target):
        request_id = next(self._request_counter)

        loop = asyncio.get_running_loop()
        future = loop.create_future()
        self._pending_requests[request_id] = future

        self.send({"type": "save-image", "target": target, "requestId": request_id})
        return future

    async def _save_image_to_file(self, path, target):
        data = await self._request_screenshot(target)
        with open(path, "wb") as f:
            f.write(data)

    async def _process_structure_sequence(self, indices, settings=None):
        request_id = next(self._request_counter)

        loop = asyncio.get_running_loop()
        future = loop.create_future()

        self._pending_requests[request_id] = {
            "future": future,
            "results": [None] * len(indices),
            "type": "sequence",
        }

        msg = {
            "type": "save-structure-sequence",
            "indices": indices,
            "requestId": request_id,
        }
        if settings is not None:
            msg["settings"] = settings

        self.send(msg)

        return await future

    def save(self, path):
        """
        Save the dataset displayed by this widget as JSON to the given ``path``.
        If ``path`` ends with ``.gz``, the file is written as gzip compressed
        JSON string.

        :param str path: where to save the dataset.
        """
        if path.endswith(".gz"):
            file = gzip.open(path, "w", 9)
        else:
            file = open(path, "wb")

        # update the settings in the data to the latest value
        data = json.loads(self.value)
        data["settings"] = self.settings
        if len(self._structures_cache) > 0:
            # there are cached structures, add them to the dump
            data["structures"] = list(self._structures_cache.values())

        file.write(json.dumps(data).encode("utf8"))
        file.close()

    def _handle_js_msg(self, _, content, buffers):
        """Handle custom messages sent from the JS widget.
        The pattern is that a message is sent from another
        function, and the response is collected here, handling
        a successful response (`msg_type="base-mst-result"`)
        or and error (`msg_type="base-mst-result"`). See below
        for a brief explanation of the different types of messages.

        Currently implemented messages:

        * "load-structure": Initiated on the JS side, used to
            retrieve an atomic geometry from disk or from a
            python-side cache.

            {"type": "load-structure",
             "requestId": int,
             "data": "filename-or-id"}

        * "save-image": Initiated on the Python side, used to get
            a snapshot of the map or of the active structure

            {"type": "save-image",
             "requestId": int,
             "target": "map" | "structure"}

        * "save-structure-sequence": Initiated on the Python side, used to
            generate snapshots for several atomic configurations.
            Should provide indices (in the same format as for `selected_ids`)
            and optionally structure settings as a list of dicts.

            {"type": "save-structure-sequence",
             "requestId": int,
             "indices": [{"structure": id1[, "atom": id2]}, ...],
             ["settings": [settings_1_dict, ...]]
             }

        Expected messages:

        - {"type": "load-structure",
           "requestId": int,
           "filename": "relative/path.json[.gz]"}
        """
        msg_type = content.get("type", None)
        request_id = content.get("requestId")

        if msg_type == "save-image-result":
            pending = self._pending_requests.pop(request_id, None)
            if pending and isinstance(pending, asyncio.Future):
                data = content.get("data")
                try:
                    _header, encoded = data.split(",", 1)
                    data = base64.b64decode(encoded)
                    pending.set_result(data)
                except Exception as e:
                    pending.set_exception(e)

        elif msg_type == "save-image-error":
            pending = self._pending_requests.pop(request_id, None)
            if pending and isinstance(pending, asyncio.Future):
                pending.set_exception(RuntimeError(content.get("error")))

        elif msg_type == "save-structure-sequence-result":
            pending = self._pending_requests.get(request_id)
            if pending and isinstance(pending, dict):
                step = content.get("step")
                data = content.get("data")
                results = pending["results"]

                try:
                    _header, encoded = data.split(",", 1)
                    decoded = base64.b64decode(encoded)
                    results[step] = decoded
                except Exception as e:
                    print(f"Error decoding frame {step}: {e}")

        elif msg_type == "save-structure-sequence-done":
            pending = self._pending_requests.pop(request_id, None)
            if pending and isinstance(pending, dict):
                pending["future"].set_result(pending["results"])

        elif msg_type == "save-structure-sequence-error":
            pending = self._pending_requests.pop(request_id, None)
            if pending and isinstance(pending, dict):
                pending["future"].set_exception(RuntimeError(content.get("error")))

        if msg_type == "load-structure":
            # Reads a structure file and sends it back to the JS side
            data = content.get("data")

            if data is None:
                self.send(
                    {
                        "type": "load-structure-error",
                        "requestId": request_id,
                        "error": "missing filename in load-structure message",
                    }
                )
                return
            if data in self._structures_cache:
                # structure is cached
                self.send(
                    {
                        "type": "load-structure-result",
                        "requestId": request_id,
                        "structure": self._structures_cache[data],
                        "data": data,  # also sends back data to cross-check
                    }
                )
            else:
                # structure is an external file
                try:
                    # read structure from file (including relative path)
                    path = Path(data)

                    # Allow gzip-compressed structure files as well
                    if str(path).endswith(".gz"):
                        with gzip.open(path, "rt") as f:
                            structure = json.load(f)
                    else:
                        with path.open("rt") as f:
                            structure = json.load(f)
                    # `structure` must be a plain JSON-serialisable object that matches
                    # chemiscope's `Structure` interface on the JS side.
                    self.send(
                        {
                            "type": "load-structure-result",
                            "requestId": request_id,
                            "structure": structure,
                            "data": data,  # also sends back data to cross-check
                        }
                    )
                except Exception as exc:
                    self.send(
                        {
                            "type": "load-structure-error",
                            "requestId": request_id,
                            "error": f"failed to load structure from '{data}': {exc}",
                        }
                    )

    def __repr__(self, max_length=64):
        # string representation of the chemiscope widget, outputs that are too large
        class_name = self.__class__.__name__
        string_repr = ""

        # loops over metadata, structures, settings, ...
        for key, value in json.loads(self.value).items():
            value_repr = repr(value)
            truncated_repr = (
                (value_repr[:max_length] + "...}")
                if len(value_repr) > max_length
                else value_repr
            )
            string_repr += f"{key}={truncated_repr}, "

        return f"<{class_name}({string_repr})>"

    def _repr_html_(self):
        if _is_running_in_notebook():
            return self.__repr__()
        else:
            return ""


@ipywidgets.register
class ChemiscopeWidget(ChemiscopeWidgetBase):
    _view_name = Unicode("ChemiscopeView").tag(sync=True)

    def __init__(self, data, has_metadata, warning_timeout, cache_structures):
        super().__init__(data, has_metadata, warning_timeout, cache_structures)


@ipywidgets.register
class StructureWidget(ChemiscopeWidgetBase):
    _view_name = Unicode("StructureView").tag(sync=True)

    def __init__(self, data, has_metadata, warning_timeout, cache_structures):
        super().__init__(data, has_metadata, warning_timeout, cache_structures)


@ipywidgets.register
class MapWidget(ChemiscopeWidgetBase):
    _view_name = Unicode("MapView").tag(sync=True)

    def __init__(self, data, has_metadata, warning_timeout, cache_structures):
        super().__init__(data, has_metadata, warning_timeout, cache_structures)


def show_input(
    path, *, settings=None, mode="default", warning_timeout=10000, cache_structures=True
):
    """
    Loads and shows the chemiscope input in ``path``.

    If ``path`` ends with ``.gz``, the file is loaded as a gzip compressed JSON string.
    If ``path`` is a file-like object, it is read as JSON input.

    :param str | Path | file-like path: load the chemiscope input from this path or
        file-like object

    :param dict settings: override the default settings in the input
    :param str mode: widget mode, either ``default``, ``structure`` or ``map``.
    :param float warning_timeout: timeout (in ms) for warnings. Set to a negative value
        to disable warnings, and to zero to make them persistent.
    :param bool cache_structures: whether to cache structure data on the Python side
        to reduce the JScript memory footprint

    .. code-block:: python

        import chemiscope

        widget = chemiscope.show_input("dataset.json")

        # or

        with open("dataset.json", "r") as f:
            widget = chemiscope.show_input(f)

    ..
    """

    if mode not in ["default", "structure", "map"]:
        raise ValueError(
            f"invalid mode '{mode}' in chemiscope.show_input, expected one of "
            "'default', 'structure' or 'map'"
        )

    if mode == "default":
        widget_class = ChemiscopeWidget

    elif mode == "structure":
        widget_class = StructureWidget

    elif mode == "map":
        widget_class = MapWidget

    if isinstance(path, Path):
        path = str(path)

    if isinstance(path, str):
        if path.endswith(".gz"):
            with gzip.open(path, "rt") as f:
                dict_input = json.load(f)
        elif path.endswith(".json"):
            with open(path, "rt") as f:
                dict_input = json.load(f)
        else:
            raise ValueError(
                "invalid file format in chemiscope.show_input, expected .json or "
                ".json.gz"
            )
    else:
        dict_input = json.load(path)

    try:
        metadata = dict_input["meta"]

        if metadata == {"name": " "}:
            has_metadata = False
        else:
            has_metadata = True

    except KeyError:
        raise ValueError("missing metadata in chemiscope.show_input")

    if settings is not None:
        if not isinstance(settings, dict):
            raise ValueError(
                f"expected 'settings' to be a dict, got {type(settings)} instead"
            )
        if "settings" not in dict_input:
            dict_input["settings"] = {}
        dict_input["settings"].update(settings)

    return widget_class(
        dict_input,
        has_metadata=has_metadata,
        warning_timeout=warning_timeout,
        cache_structures=cache_structures,
    )


def show(
    structures=None,
    *,
    properties=None,
    metadata=None,
    environments=None,
    shapes=None,
    settings=None,
    mode="default",
    warning_timeout=10000,
    cache_structures=True,
    # for backward compatibility
    frames=None,
    meta=None,
):
    """
    Show the dataset defined by the given ``structures`` and ``properties`` (optionally
    ``metadata``, ``environments`` and ``shapes`` as well) using an embedded chemiscope
    visualizer inside a Jupyter notebook. These parameters have the same meaning as in
    the :py:func:`chemiscope.create_input` function.

    The ``mode`` keyword also allows overriding the default two-panels visualization to
    show only a structure panel (``mode = "structure"``) or the map panel (``mode =
    "map"``). These modes also make it possible to view a dataset for which properties
    (or structures) are not available. The widget displays warning messages, that
    disappear after the specified ``warning_timeout`` (in ms). Set to a negative value
    to disable warnings, and to zero to make them persistent. ``cache_structures`` is a
    flag determining whether to cache structure data on the Python side to reduce the
    JScript memory footprint.

    When inside a jupyter notebook, the returned object will create a new chemiscope
    visualizer displaying the dataset. The object exposes a ``settings`` traitlet, that
    allows to modify the visualization options (possibly even linking the parameters to
    another widget). Printing the value of the ``settings`` property is also a good way
    to see a full list of the available options.

    The returned object also have a ``save`` function that can be used to save the
    dataset to a ``.json`` or ``.json.gz`` file to load it in the main website later.
    The visualization options will be those used in the active widget, so this is also a
    good way to tweak the appearance of the visualization before saving it.

    .. code-block:: python

        import chemiscope
        from sklearn.decomposition import PCA
        import ase.io

        pca = PCA(n_components=3)

        structures = ase.io.read(...)
        properties = {
            "PCA": pca.fit_transform(some_data),
        }

        widget = chemiscope.show(structures, properties)
        # display the dataset in a chemiscope visualizer inside the notebook
        widget
        # ...

        # NB: due to how traitlet work, you should always set the value of
        # the `settings` property. Only the properties that are explicitly
        # indicated will be modified.
        widget.settings = {"map": {"symbol": "tag"}}
        widget.settings["map"]["symbol"] = "tag"  # << does nothing!

        # Save the file for later use
        widget.save("dataset.json")

    .. _ase.Atoms: https://wiki.fysik.dtu.dk/ase/ase/atoms.html
    """
    if frames is not None:
        warnings.warn(
            "`frames` argument is deprecated, use `structures` instead",
            stacklevel=2,
        )
        if structures is not None:
            raise ValueError("cannot use both `structures` and `frames` arguments")

        structures = frames

    if meta is not None:
        warnings.warn(
            "`meta` argument is deprecated, use `metadata` instead",
            stacklevel=2,
        )
        if metadata is not None:
            raise ValueError("cannot use both `metadata` and `meta` arguments")

        metadata = meta

    if not (_is_running_in_notebook() or _is_running_in_sphinx_gallery()):
        warnings.warn(
            "chemiscope.show only works in a jupyter notebook or a sphinx build",
            stacklevel=2,
        )

    has_metadata = metadata is not None
    if not has_metadata:
        metadata = {"name": " "}

    if mode == "default":
        widget_class = ChemiscopeWidget

    elif mode == "structure":
        if properties is None:
            properties = {}

        widget_class = StructureWidget

    elif mode == "map":
        if properties is None:
            properties = {}

        widget_class = MapWidget

    else:
        raise ValueError(
            f"invalid mode '{mode}' in chemiscope.show, expected one of "
            "'default', 'structure' or 'map'"
        )

    dict_input = create_input(
        structures=structures,
        properties=properties,
        metadata=metadata,
        environments=environments,
        shapes=shapes,
        settings=settings,
    )

    if mode == "structure" and "index" not in dict_input["properties"]:
        dict_input["properties"]["index"] = {
            "target": "structure",
            "values": list(range(len(dict_input["structures"]))),
        }

    if mode != "structure":
        # if there is a map, we need two properties, otherwise there will be no map
        # and error is only visible in the console
        if len(dict_input["properties"]) < 2:
            raise ValueError("Need at least two properties to visualize a map widget")

    return widget_class(
        dict_input,
        has_metadata=has_metadata,
        warning_timeout=warning_timeout,
        cache_structures=cache_structures,
    )


def _is_running_in_notebook():
    """
    Check whether the python interpreter is running for a jupyter notebook or
    not. Taken from https://stackoverflow.com/a/39662359/4692076.
    """

    # apparently get_ipython is lost when this gets called from a callback of
    # an ipython widget. See https://github.com/jupyter/jupyter/issues/299
    try:
        from IPython import get_ipython
    except ImportError:
        return False

    try:
        shell = get_ipython().__class__.__name__
        # ZMQInteractiveShell is the standard Jupyter Kernel
        # Interpreter is used by pyiodide
        if shell in ["ZMQInteractiveShell", "Interpreter"]:
            return True
        elif shell == "TerminalInteractiveShell":
            return False
        else:
            return False
    except NameError:
        return False


def _is_running_in_sphinx_gallery():
    """
    Returns true if a file is being executed by sphinx-gallery.
    """

    # This is very hacky: we are relying on the fact that `sphinx_gallery.gen_gallery`
    # should only be imported when actually generating the gallery, and is not imported
    # otherwise.
    try:
        import sphinx_gallery

        sphinx_build = hasattr(sphinx_gallery, "gen_gallery")
    except ImportError:
        sphinx_build = False

    return sphinx_build



import warnings

import numpy as np

from ._shapes import arrow_from_vector, ellipsoid_from_tensor


try:
    import ase

    HAVE_ASE = True
except ImportError:
    HAVE_ASE = False


def _ase_valid_structures(structures):
    try:
        structures_list = list(structures)
    except TypeError:
        return [], False

    if HAVE_ASE and isinstance(structures_list[0], ase.Atoms):
        for structure in structures_list:
            assert isinstance(structure, ase.Atoms)
        return structures_list, True
    elif HAVE_ASE and isinstance(structures_list[0], ase.Atom):
        # deal with the user passing a single structure
        return [structures], True
    else:
        return structures_list, False


def _ase_normalize_value(value):
    if isinstance(value, np.ndarray):
        return value.tolist()
    elif isinstance(value, (np.int32, np.int64)):
        return int(value)
    elif isinstance(value, (np.float32, np.float64)):
        return float(value)
    else:
        return value


def _ase_get_atom_properties(structures):
    IGNORED_ASE_ARRAYS = ["positions", "numbers", "center_atoms_mask"]
    # extract the set of common properties between all structures
    all_properties = {}
    extra = set()

    arrays = structures[0].arrays.copy()
    # workaround for ase >= 3.23
    if structures[0].calc is not None and "forces" in structures[0].calc.results:
        arrays["forces"] = structures[0].calc.results["forces"]

    for name in arrays.keys():
        if name in IGNORED_ASE_ARRAYS:
            continue
        all_properties[name] = [arrays[name]]

    for structure in structures[1:]:
        arrays = structure.arrays.copy()
        # workaround for ase >= 3.23
        if structure.calc is not None and "forces" in structure.calc.results:
            arrays["forces"] = structure.calc.results["forces"]

        for name in arrays.keys():
            if name in IGNORED_ASE_ARRAYS:
                continue

            if name in all_properties:
                all_properties[name].append(arrays[name])
            else:
                extra.add(name)

        for name in list(all_properties.keys()):
            if name not in arrays.keys():
                all_properties.pop(name, None)
                extra.add(name)

    if len(extra) != 0:
        warnings.warn(
            "the following atomic properties are only defined for a subset "
            f"of structures: {list(sorted(extra))}; they will be ignored",
            stacklevel=2,
        )

    return all_properties


def _ase_get_structure_properties(structures):
    # extract the set of common properties between all structures
    all_properties = {}
    extra = set()

    info = structures[0].info.copy()
    # workaround for ase >= 3.23
    if structures[0].calc is not None and "energy" in structures[0].calc.results:
        info["energy"] = structures[0].calc.results["energy"]

    for name in info.keys():
        all_properties[name] = [_ase_normalize_value(info[name])]

    for structure in structures[1:]:
        info = structure.info.copy()
        # workaround for ase >= 3.23
        if structure.calc is not None and "energy" in structure.calc.results:
            info["energy"] = structure.calc.results["energy"]

        for name in info.keys():
            if name in all_properties:
                all_properties[name].append(_ase_normalize_value(info[name]))
            else:
                extra.add(name)

        for name in list(all_properties.keys()):
            if name not in info.keys():
                all_properties.pop(name, None)
                extra.add(name)

    if len(extra) != 0:
        warnings.warn(
            "the following structure properties are only defined for a subset "
            f"of structures: {list(sorted(extra))}; they will be ignored",
            stacklevel=2,
        )

    # ensures that if a property is a mix of strings and numbers, everything
    # is converted to string (as these should be categorical properties)
    for name in all_properties.keys():
        property = all_properties[name]
        if any(isinstance(x, str) for x in property):
            all_properties[name] = [str(x) for x in property]

    return all_properties


def _ase_atom_properties(structures, only=None, atoms_mask=None):
    all_properties = _ase_get_atom_properties(structures)
    if only is None:
        selected = all_properties
    else:
        selected = {}
        for name in only:
            if name in all_properties.keys():
                selected[name] = all_properties[name]

    # create property in the format expected by create_input
    properties = {
        name: {"target": "atom", "values": np.concatenate(value)}
        for name, value in selected.items()
    }

    _remove_invalid_properties(properties, "ASE")

    if atoms_mask is not None:
        # only include values for requested atoms
        for property in properties.values():
            property["values"] = property["values"][atoms_mask]

    return properties


def _ase_structure_properties(structures, only=None):
    all_properties = _ase_get_structure_properties(structures)
    if only is None:
        selected = all_properties
    else:
        selected = {}
        for name in only:
            if name in all_properties.keys():
                selected[name] = all_properties[name]

    # create property in the format expected by create_input
    properties = {
        name: {"target": "structure", "values": value}
        for name, value in selected.items()
    }

    _remove_invalid_properties(properties, "ASE")

    return properties


def _ase_extract_properties(structures, only=None, environments=None):
    """implementation of ``extract_properties`` for ASE"""
    properties = _ase_structure_properties(structures, only)

    if environments is not None:
        atoms_mask = [[False] * len(f) for f in structures]
        for structure, center, _ in environments:
            atoms_mask[structure][center] = True

        atoms_mask = np.concatenate(atoms_mask)
    else:
        atoms_mask = None

    atom_properties = _ase_atom_properties(structures, only, atoms_mask)

    for name, values in atom_properties.items():
        if name in properties:
            warnings.warn(
                f"a property named '{name}' is defined for both atoms and structures, "
                "the atom one will be ignored",
                stacklevel=2,
            )
        else:
            properties[name] = values

    return properties


def _ase_all_atomic_environments(structures, cutoff):
    "Extract all atomic environments out of a set of ASE Atoms objects"
    environments = []
    for structure_i, structure in enumerate(structures):
        for atom_i in range(len(structure)):
            environments.append((structure_i, atom_i, cutoff))
    return environments


def _ase_to_json(structure):
    """Implementation of structures_to_json for ase.Atoms"""
    data = {}
    data["size"] = len(structure)
    data["names"] = list(structure.symbols)
    data["x"] = structure.positions[:, 0].tolist()
    data["y"] = structure.positions[:, 1].tolist()
    data["z"] = structure.positions[:, 2].tolist()

    if (structure.cell.lengths() != [0.0, 0.0, 0.0]).all():
        data["cell"] = list(np.concatenate(structure.cell))

    return data


def _remove_invalid_properties(properties, origin):
    """
    Remove invalid properties from the ``properties`` dictionary. ``origin`` is
    used in error messages as the property origin
    """
    to_remove = set()

    for name, property in properties.items():
        values = property["values"]

        for value in values:
            # check type consistency
            if not _is_convertible_to_property(value):
                warnings.warn(
                    f"value '{value}' of type '{type(value)}' for the '{name}' "
                    f"property from {origin} is not convertible to float, array or "
                    "string, this property will be ignored.",
                    stacklevel=2,
                )
                to_remove.add(name)
                break

        if (
            name in to_remove
            or not hasattr(values[0], "__len__")
            or isinstance(values[0], str)
        ):
            continue

        # check length consistency
        lengths = [len(v) for v in property["values"]]
        if len(set(lengths)) != 1:
            to_remove.add(name)
            warnings.warn(
                f"values of the property '{name}' have inconsistent length across "
                "different structures, it will be ignored",
                stacklevel=2,
            )

    for name in to_remove:
        del properties[name]


def _is_convertible_to_property(value):
    """
    Check whether a value is convertible to a chemiscope property, i.e. if it is
    a string or something convertible to float.
    """
    if isinstance(value, (bytes, str)) or (
        (isinstance(value, list) or isinstance(value, np.ndarray))
        and isinstance(value[0], str)
    ):
        # string types
        return True
    else:
        # everything convertible to float
        try:
            float(value)
            return True
        except Exception:
            try:
                np.array(value, dtype=float)
                return True
            except Exception:
                return False


def _extract_key_from_ase(structure, key, target=None):
    """
    Extract a property from either the atomic array of info fields of an ``ase.Atoms``
    structure. Defaults to atoms if no target is specified, and also returns the actual
    target it picked the key from.
    """

    arrays = structure.arrays.copy()
    info = structure.info.copy()
    # ase >= 3.23 workaround
    if structure.calc is not None and "forces" in structure.calc.results:
        arrays["forces"] = structure.calc.results["forces"]
    if structure.calc is not None and "energy" in structure.calc.results:
        info["energy"] = structure.calc.results["energy"]

    if target is None:
        if key in arrays:  # defaults to atom target
            target = "atom"
        elif key in info:
            target = "structure"
        else:
            raise IndexError(
                f"Key {key} not found in neither `Atoms.arrays` or `Atoms.info`"
            )

    if target == "atom":
        try:
            values = arrays[key]
        except IndexError:
            raise IndexError(f"Key {key} not found in `Atoms.arrays`")
    if target == "structure":
        try:
            values = info[key]
        except IndexError:
            raise IndexError(f"Key {key} not found in `Atoms.info`")

    return values, target


def ase_vectors_to_arrows(
    structures=None,
    key=None,
    target=None,
    *,
    scale=1.0,
    radius=0.1,
    head_radius_scale=1.75,
    head_length_scale=2.0,
    frames=None,
):
    """
    Extract a vectorial atom property from a list of ``ase.Atoms``, and returns
    a list of arrow shapes. Besides the specific parameters it also accepts the same
    parameters as :py:func:`arrow_from_vector`, which are used to define the style of
    the arrows.

    :param structures: list of ASE Atoms
    :param key: name of the ASE atom property. Should contain three components
       corresponding to x,y,z
    :param target: whether the properties should be associated with the entire
       structure, or each atom (``structure`` or ``atom``). defaults to autodetection
    """
    if frames is not None:
        warnings.warn(
            "`frames` parameter is deprecated, please use `structures` instead",
            stacklevel=2,
        )

        if structures is not None:
            raise ValueError("cannot use both `structures` and `frames` parameters")

        structures = frames

    if key is None:
        # this is only required while we keep support for the deprecated `frames`
        # parameter. We can remove the default None key when we remove `frames`
        raise ValueError("`key` parameter must be provided")

    vectors = []

    # set shape parameters globally to reduce size of the arrow list
    globs = {}
    globs["baseRadius"] = radius
    globs["headRadius"] = globs["baseRadius"] * head_radius_scale
    globs["headLength"] = globs["baseRadius"] * head_length_scale

    for structure in structures:
        values, target = _extract_key_from_ase(structure, key, target)
        if target == "structure":
            values = values.reshape(1, -1)
        if len(values.shape) != 2 or values.shape[1] != 3:
            raise ValueError(
                f"Property array {key} has not the shape of a list of 3-vectors"
            )

        # makes a list of arrows to visualize the property
        vectors = vectors + [
            arrow_from_vector(
                v,
                scale=scale,
                radius=None,
                head_radius_scale=None,
                head_length_scale=None,
            )
            for v in values
        ]

    if target == "atom":
        return {"kind": "arrow", "parameters": {"global": globs, "atom": vectors}}
    else:
        return {"kind": "arrow", "parameters": {"global": globs, "structure": vectors}}


def ase_tensors_to_ellipsoids(
    structures=None,
    key=None,
    target=None,
    *,
    scale=1.0,
    force_positive=False,
    frames=None,
):
    """
    Extract a 2-tensor atom property from a list of ``ase.Atoms``, and returns a list of
    ellipsoids shapes. Besides the specific parameters it also accepts the same
    parameters as :py:func:`ellipsoid_from_tensor`, which are used to draw the shapes

    :param structures: list of ASE Atoms
    :param key: name of the ASE atom property. Should contain nine components
       corresponding to xx,xy,xz,yx,yy,yz,zx,zy,zz or six components corresponding to
       xx,yy,zz,xy,xz,yz
    :param target: whether the properties should be associated with the entire
       structure, or each atom (``structure`` or ``atom``). defaults to autodetection

    :param scale: see :py:func:`ellipsoid_from_tensor`
    :param force_positive: see :py:func:`ellipsoid_from_tensor`
    """

    if frames is not None:
        warnings.warn(
            "`frames` parameter is deprecated, please use `structures` instead",
            stacklevel=2,
        )

        if structures is not None:
            raise ValueError("cannot use both `structures` and `frames` parameters")

        structures = frames

    if key is None:
        # this is only required while we keep support for the deprecated `frames`
        # parameter. We can remove the default None key when we remove `frames`
        raise ValueError("`key` parameter must be provided")

    tensors = []

    for structure in structures:
        values, target = _extract_key_from_ase(structure, key, target)
        if target == "structure":
            values = values.reshape(1, -1)
        if len(values.shape) != 2 or (values.shape[1] != 6 and values.shape[1] != 9):
            raise ValueError(
                f"Property array {key} has not the shape of a list of 6 or 9-vectors"
            )

        # makes a list of ellipsoids to visualize the property
        tensors = tensors + [
            ellipsoid_from_tensor(v, scale=scale, force_positive=force_positive)
            for v in values
        ]

    if target == "atom":
        return dict(kind="ellipsoid", parameters={"global": {}, "atom": tensors})
    else:
        return dict(kind="ellipsoid", parameters={"global": {}, "structure": tensors})


# Required parameters from different kinds of shapes
SHAPE_PARAMETERS = {
    "ellipsoid": "semiaxes",
    "sphere": "radius",
}


def _extract_lammps_shapes(structure, key):
    if key in structure.info:
        if structure.info[key] not in SHAPE_PARAMETERS:
            raise KeyError(
                "The currently-supported shape in `extract_lammps_shapes_from_ase` are "
                f"{list(SHAPE_PARAMETERS.keys())}, received '{structure.info[key]}'"
            )

        shape = _get_shape_params(key, structure.info[key], structure.info)
        if "orientation" in structure.arrays:
            return {
                key: [
                    {**shape, "orientation": list(o)}
                    for o in structure.arrays["orientation"]
                ]
            }
        else:
            return {key: [shape for _ in structure]}

    elif key in structure.arrays:
        shapes = []
        for atom_i, shape_key in enumerate(structure.arrays[key]):
            if shape_key not in SHAPE_PARAMETERS:
                raise KeyError(
                    "The currently-supported shape types are {}, received {}.".format(
                        ", ".join(SHAPE_PARAMETERS.keys()), shape_key
                    )
                )
            shape = _get_shape_params_atom(
                key,
                shape_key,
                structure.arrays,
                atom_i,
            )

            if "orientation" in structure.arrays:
                shape["orientation"] = list(structure.arrays["orientation"][atom_i])

            shapes.append(shape)

        return {key: shapes}


def _get_shape_params(prefix, shape_kind, dictionary):
    shape = {"kind": shape_kind}
    parameter = SHAPE_PARAMETERS[shape_kind]
    try:
        shape[parameter] = dictionary[f"{prefix}_{parameter}"]
    except KeyError:
        raise KeyError(
            f"Missing required parameter '{prefix}_{parameter}' for "
            f"'{shape_kind}' shape"
        )

    return shape


def _get_shape_params_atom(prefix, shape_kind, dictionary, atom_i):
    """Extract shape parameters for a single atom"""

    shape = {"kind": shape_kind}
    parameter = SHAPE_PARAMETERS[shape_kind]
    try:
        shape[parameter] = dictionary[f"{prefix}_{parameter}"][atom_i]
    except KeyError:
        raise KeyError(
            f"Missing required parameter '{prefix}_{parameter}' for "
            f"'{shape_kind}' shape"
        )

    return shape



# -*- coding: utf-8 -*-
import warnings

from ._ase import (  # noqa: F401
    _ase_all_atomic_environments,
    _ase_extract_properties,
    _ase_to_json,
    _ase_valid_structures,
    ase_tensors_to_ellipsoids,
    ase_vectors_to_arrows,
)
from ._mda import (  # noqa: F401
    _mda_all_atomic_environments,
    _mda_extract_properties,
    _mda_to_json,
    _mda_valid_structures,
)
from ._shapes import (  # noqa: F401
    arrow_from_vector,
    ellipsoid_from_tensor,
)
from ._stk import (  # noqa: F401
    _stk_all_atomic_environments,
    _stk_to_json,
    _stk_valid_structures,
)


def _chemiscope_valid_structures(structures):
    """
    Check if the given structures are already in chemiscope format.

    :param structures: iterable over structures
    :return: tuple (structures as list, boolean indicating if structures are valid)
    """

    if not hasattr(structures, "__iter__"):
        return structures, False

    first_structure = structures[0]
    if not isinstance(first_structure, dict):
        return structures, False

    required_keys = {"size", "names", "x", "y", "z"}
    if "size" not in first_structure:
        return structures, False

    for s in structures:
        assert isinstance(s, dict), (
            "inconsistent structure types: "
            f"expected dict but got {s.__class__.__name__}"
        )

        # this is an external structure, so we skip the detailed checks
        if "data" not in s:
            assert required_keys.issubset(set(s.keys())), (
                "invalid chemiscope structure: "
                f"missing keys {required_keys - set(s.keys())}"
            )

    return structures, True


def _guess_adapter(structures):
    """
    Guess which adapter to use for the given structures. This function return the
    structures as a list and a string describing which adapter should be used.
    """

    chemiscope_structures, use_chemiscope = _chemiscope_valid_structures(structures)
    if use_chemiscope:
        return chemiscope_structures, "chemiscope"

    ase_structures, use_ase = _ase_valid_structures(structures)
    if use_ase:
        return ase_structures, "ASE"

    stk_structures, use_stk = _stk_valid_structures(structures)
    if use_stk:
        return stk_structures, "stk"

    mda_structures, use_mda = _mda_valid_structures(structures)
    if use_mda:
        return mda_structures, "mda"

    raise Exception(f"unknown structure type: '{structures[0].__class__.__name__}'")


def structures_to_json(structures):
    """
    Convert the given ``structures`` to the JSON structure used by chemiscope.

    This function is a shim calling specialized implementations for all the
    supported structure types. Currently supported structures types include
    chemiscope-compatible dicts, ``ase.Atoms``, ``stk.BuildingBlocks``, and
    ``MDAnalysis.AtomGroup`` objects.

    :param structures: iterable over structures
    """
    structures, adapter = _guess_adapter(structures)

    if adapter == "chemiscope":
        json_data = structures
    elif adapter == "ASE":
        json_data = [_ase_to_json(s) for s in structures]
    elif adapter == "stk":
        json_data = [_stk_to_json(s) for s in structures]
    elif adapter == "mda":
        # Be careful of the lazy loading of `structures.atoms`, which is updated during
        # the iteration of the trajectory
        json_data = [_mda_to_json(structures) for _ in structures.universe.trajectory]
    else:
        raise Exception("reached unreachable code")

    return json_data


def extract_properties(structures=None, only=None, *, environments=None, frames=None):
    """
    Extract properties defined in the ``structures`` in a chemiscope-compatible format.

    :param structures: iterable over structures
    :param only: optional, list of strings. If not ``None``, only properties with a name
                from this list are included in the output.
    :param environments: optional, list of environments (described as ``(structure id,
        center id, cutoff)``) to include when extracting the atomic properties.
    """
    if frames is not None:
        warnings.warn(
            "`frames` argument is deprecated, use `structures` instead",
            stacklevel=2,
        )
        if structures is not None:
            raise ValueError("cannot use both `structures` and `frames` arguments")

        structures = frames

    structures, adapter = _guess_adapter(structures)

    if adapter == "ASE":
        return _ase_extract_properties(structures, only, environments)

    elif adapter == "mda":
        return _mda_extract_properties(structures, only, environments)

    elif adapter == "stk":
        raise RuntimeError(
            "stk structures do not contain properties, you must manually provide them"
        )

    else:
        raise Exception("reached unreachable code")


def all_atomic_environments(structures=None, cutoff=3.5, *, frames=None):
    """
    Generate a list of environments containing all the atoms in the given
    ``structures``. The optional spherical ``cutoff`` radius is used to display the
    environments in chemiscope.

    :param structures: iterable over structures
    :param float cutoff: spherical cutoff radius used when displaying the
                         environments
    """
    if frames is not None:
        import warnings

        warnings.warn(
            "`frames` argument is deprecated, use `structures` instead",
            stacklevel=2,
        )
        if structures is not None:
            raise ValueError("cannot use both `structures` and `frames` arguments")

        structures = frames

    structures, adapter = _guess_adapter(structures)

    if adapter == "ASE":
        return _ase_all_atomic_environments(structures, cutoff)
    elif adapter == "stk":
        return _stk_all_atomic_environments(structures, cutoff)
    elif adapter == "mda":
        return _mda_all_atomic_environments(structures, cutoff)
    else:
        raise Exception("reached unreachable code")



import os
from typing import Any, Callable, Mapping, Optional


_component_func = None


def viewer(
    dataset: Mapping[str, Any],
    *,
    settings: Optional[dict] = None,
    mode: str = "default",
    no_info_panel: bool = False,
    width: str | int = "stretch",
    height: int = 550,
    key: str = "chemiscope_viewer",
    selected_index: Optional[int] = None,
    on_select: Optional[Callable[[Optional[int]], Any]] = None,
    on_settings_change: Optional[Callable[[dict], Any]] = None,
) -> Any:
    """
    Render a Chemiscope viewer inside a Streamlit app.

    :param dict dataset: Chemiscope dataset (same dict generated by
        :py:func:`chemiscope.create_input`).
    :param dict settings: Optional settings dictionary to configure the viewer
        appearance and behavior, as generated e.g. by
        :py:func:`chemiscope.quick_settings`.
    :param str mode: Visualization mode: "default", "structure", or "map".
    :param bool no_info_panel: Hide the info panel, disabling in-widget selection
        for ``structure`` mode
    :param str|int width: Width in pixels or "stretch" for full width (default).
    :param int height: Height in pixels for the viewer (or None to auto-size).
    :param str key: Streamlit widget key.
    :param int selected_index: Structure index to select. Can also be used
        to update the selection from outside the component if set to a streamlit
        state variable.
    :param callable on_select: Optional callback function called when the
        structure selection changes. Receives the new selected index or None
        if unselected.
    :param callable on_settings_change: Called when settings change, settings
        dict as argument.
    """
    global _component_func

    try:
        import streamlit as st
        import streamlit.components.v1 as components
    except ImportError:
        raise ImportError(
            "Streamlit is required to use chemiscope.streamlit.viewer. "
            "Install it with: pip install 'chemiscope[streamlit]'"
        )

    if _component_func is None:
        _component_func = components.declare_component(
            "chemiscope_viewer",
            path=os.path.relpath(os.path.dirname(__file__)),
        )

    if mode not in ("default", "structure", "map"):
        raise ValueError(
            f"Invalid mode '{mode}', expected 'default', 'structure', or 'map'"
        )

    has_structures = "structures" in dataset and len(dataset["structures"]) > 0
    has_properties = "properties" in dataset and len(dataset["properties"]) > 0
    if mode == "structure" and not has_structures:
        raise ValueError("Cannot show structure: dataset has no structures")
    if mode == "map" and (not has_properties or len(dataset["properties"]) < 2):
        raise ValueError("Map mode requires at least two properties")

    state_key = f"{key}_state"
    if state_key not in st.session_state:
        st.session_state[state_key] = {
            "selected_index": selected_index,
            "settings": settings or {},
            "last_update": None,
        }

    def on_change():
        current_state = st.session_state[state_key]
        component_value = st.session_state[key]

        if component_value is None:
            return

        new_selection = component_value.get("selected_id")
        new_settings = component_value.get("settings", {})

        selection_changed = new_selection != current_state["selected_index"]
        settings_changed = new_settings != current_state["settings"]

        if selection_changed:
            current_state["selected_index"] = new_selection
            if on_select:
                on_select(new_selection)

        if settings_changed:
            current_state["settings"] = new_settings
            if on_settings_change:
                on_settings_change(new_settings)

        if selection_changed or settings_changed:
            current_state["last_update"] = "component"

    current_state = st.session_state[state_key]

    external_selection_changed = (
        selected_index is not None
        and selected_index != current_state["selected_index"]
        and current_state["last_update"] != "streamlit"
    )

    external_settings_changed = (
        settings is not None
        and settings != current_state["settings"]
        and current_state["last_update"] != "streamlit"
    )

    if external_selection_changed:
        current_state["selected_index"] = selected_index
        current_state["last_update"] = "streamlit"

    if external_settings_changed:
        current_state["settings"] = settings
        current_state["last_update"] = "streamlit"

    if current_state["last_update"] == "streamlit":
        current_state["last_update"] = None

    return _component_func(
        dataset=dataset,
        settings=current_state["settings"],
        mode=mode,
        no_info_panel=no_info_panel,
        key=key,
        width=width,
        height=height,
        selected_index=current_state["selected_index"],
        default=None,
        on_change=on_change,
    )



import hashlib
import os

from docutils.parsers.rst import Directive

from ._nodes import chemiscope
from ._utils import copy_external_structures, copy_file


class ChemiscopeDirective(Directive):
    """Custom RST directive to include a chemiscope visualizer in a
    sphinx documentation file. It has two options: `filepath` - the path to a chemiscope
    JSON or gzipped JSON file, relative to the path of the RST file, and `mode`, which
    can be `default`, `structure` or `map` depending on the desired type of
    visualization. `warning_timeout` specifies the time in milliseconds to display
    a warning (set it to 0 to make them persistent, and to -1 to disable).

    e.g.::

        .. chemiscope:: datasets/polarizability.json.gz
            :mode: map
            :warning_timeout: 1000

    The resulting html is the chemiscope widget wrapped in the chemiscope-sphinx.html
    template.
    """

    node_class = chemiscope
    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {"mode": str, "warning_timeout": float}
    pages_with_headers = []

    def run(self):
        # Get the source path from the document
        source = self.state.document["source"]
        source_path = os.path.dirname(source) + "/"

        # Path to the saved dataset in the .rst files folder
        dataset_rel_path = self.arguments[0].strip()
        dataset_path = source_path + dataset_rel_path

        # Ensure unique file name to avoid clashes if the same file name is used in
        # different directives.
        with open(dataset_path, "rb") as fd:
            shasum = hashlib.sha256(fd.read()).hexdigest()
        filename = f"{shasum}-{os.path.basename(dataset_path)}"

        # Copy dataset to the docs/build/html/_datasets folder
        build_file_path, rel_file_path = self.get_build_file_path(filename)
        copy_file(dataset_path, build_file_path)
        # moves files referenced by external structures in a folder with
        # a name based on the dataset file
        copy_external_structures(dataset_path, f"{build_file_path}-ext/")

        # Create the chemiscope node
        node = self.create_node(
            rel_file_path,
            include_headers=source not in ChemiscopeDirective.pages_with_headers,
        )

        # indicates that the page has already been added headers
        if source not in ChemiscopeDirective.pages_with_headers:
            ChemiscopeDirective.pages_with_headers.append(source)
        return [node]

    def get_build_file_path(self, filename):
        """
        Construct the path to the build directory

        Parameters:
        - filename (str): The name of the file

        Returns:
        - tuple: A tuple containing the build file path and the relative file path
        """

        # Get the destination folder
        outdir = self.state.document.settings.env.app.outdir
        target_dir = os.path.join(outdir, "_datasets")
        os.makedirs(target_dir, exist_ok=True)

        # Get destination paths
        build_file_path = os.path.join(target_dir, filename)

        # Get path of output dataset relative to output HTML
        env = self.state.document.settings.env
        builder = env.app.builder
        html_file_path = builder.get_outfilename(env.docname)
        html_file_dir = os.path.dirname(html_file_path)

        # Relative path *for the output files*
        rel_file_path = os.path.relpath(build_file_path, html_file_dir)
        return build_file_path, rel_file_path

    def create_node(self, rel_file_path, include_headers=True):
        """
        Create a chemiscope node with the specified file path and mode

        Parameters:
        - rel_file_path (str): The dataset path relative to the build directory

        Returns:
        - chemiscope: The created chemiscope node
        """

        node = chemiscope()
        node["filepath"] = rel_file_path
        node["mode"] = self.options.get("mode", "default")
        node["warning_timeout"] = self.options.get("warning_timeout", 4000)
        node["include_headers"] = include_headers

        self.state.nested_parse(self.content, self.content_offset, node)

        return node



# -*- coding: utf-8 -*-
"""
Generate JSON input files for the default chemiscope visualizer.
"""

import gzip
import json
import os
import warnings
from numbers import Number
from pathlib import Path

import numpy as np

from .structures import structures_to_json


def create_input(
    structures=None,
    *,
    metadata=None,
    properties=None,
    environments=None,
    settings=None,
    shapes=None,
    parameters=None,
    # for backward compatibility
    frames=None,
    meta=None,
):
    """
    Create a dictionary that can be saved to JSON using the format used by the default
    chemiscope visualizer.

    :param list structures: list of atomic structures. These can be either `chemiscope`
        compatible dictionaries, or `ase.Atoms`_, `stk.BuildingBlocks`_, and
        `MDAnalysis.AtomGroup`_ objects.

    :param dict metadata: optional metadata of the dataset, see below

    :param dict properties: optional dictionary of properties, see below

    :param list environments: optional list of ``(structure id, atom id, cutoff)``
        specifying which atoms have properties attached and how far out atom-centered
        environments should be drawn by default. Functions like
        :py:func:`all_atomic_environments` can be used to generate the list of
        environments in simple cases.

    :param dict shapes: optional dictionary of shapes to have available for display, see
        below.

    :param dict settings: optional dictionary of settings to use when displaying the
        data. Possible entries for the ``settings`` dictionary are documented in the
        chemiscope input file reference.

    :param dict parameters: optional dictionary of parameters for multidimensional
        properties, see below

    For complete format specifications, parameter types, and more examples, see `JSON
    file format <json-format>`.

    Dataset metadata
    ----------------

    The dataset metadata should be given in the ``metadata`` dictionary, the possible
    keys are:

    .. code-block:: python

        metadata = {
            # str, dataset name
            "name": "...",
            # str, dataset description
            "description": "...",
            # list of str, dataset authors, OPTIONAL
            "authors": [
                "...",
            ],
            # list of str, references for this dataset, OPTIONAL
            "references": [
                "...",
            ],
        }

    Dataset properties
    ------------------

    Properties can be added with the ``properties`` parameter. This parameter should be
    a dictionary containing one entry for each property. Properties can be extracted
    from structures with :py:func:`extract_properties`, or manually defined by the user.

    Each entry in the ``properties`` dictionary contains a ``target`` attribute
    (``'atom'`` or ``'structure'``) and a set of values. ``values`` can be a Python list
    of float or string; a 1D numpy array of numeric values; or a 2D numpy array of
    numeric values. In the later case, multiple properties will be generated along the
    second axis. For example, passing

    .. code-block:: python

        properties = {
            "cheese": {
                "target": "atom",
                "values": np.zeros((300, 4)),
                # optional: property unit
                "units": "random / fs",
                # optional: property description
                "description": "a random property for example",
            }
        }

    will generate four properties named ``cheese[1]``, ``cheese[2]``, ``cheese[3]``, and
    ``cheese[4]``, each containing 300 values.

    It is also possible to pass shortened representation of the properties, for
    instance:

    .. code-block:: python

        properties = {
            "cheese": np.zeros((300, 4)),
        }

    In this case, the type of property (structure or atom) would be deduced by comparing
    the numbers atoms and structures in the dataset to the length of provided
    list/np.ndarray.

    Multi-dimensional properties
    ----------------------------

    One can give 2D properties to be displayed as curves in the info panel by setting a
    ``parameters`` in the property, and giving the corresponding ``parameters`` values
    to this function. The previous example becomes:

    .. code-block:: python

        properties = {
            "cheese": {
                "target": "atom",
                "values": np.zeros((300, 4)),
                # optional: property unit
                "units": "random / fs",
                # optional: property description
                "description": "a random property for examples",
                "parameters": ["origin"],
            }
        }

    This input describes a 2D property ``cheese`` with 300 samples and 4 values taken by
    the ``origin`` parameter. We also need to provide the ``parameters`` values to this
    function:

    .. code-block:: python

        parameters = {
            "origin": {
                # an array of numbers containing the values of the parameter
                # the size should correspond to the second dimension
                # of the corresponding multidimensional property
                "values": [0, 1, 2, 3],
                # optional free-form description of the parameter as a string
                "name": "a short description of this parameter",
                # optional units of the values in the values array
                "units": "eV",
            }
        }

    Custom shapes
    -------------

    The ``shapes`` option should have the format ``{"<name>": shape_definition }``,
    where each shape is defined as a dictionary containing the kind of shape, and its
    parameters

    .. code-block:: python

        shapes = {
            "shape name": {
                "kind": "sphere",
                "parameters": shape_parameters,
            }
        }

    Each parameters block defines `global`, `structure` and `atom` - level parameters.

    .. code-block:: python

        parameters = {
            "global": global_parameters,
            "structure": [structure_1, structure_2, ...],
            "atom": [atom_1, atom_2, ...],
        }

    Each of these can contain some or all of the parameters associated with each shape,
    and the parameters for each shape are obtained by combining the parameters from the
    most general to the most specific, i.e., if there is a duplicate key in the
    ``global`` and ``atom`` fields, the value within the ``atom`` field will supersede
    the ``global`` field for that atom. The parameters for atom ``k`` that is part of
    structure ``j`` are obtained as

    .. code-block:: python

        global_parameters.update(structure_j).update(atom_k)

    If given, the `structure` parameters list should contain one entry per structure,
    and the `atom` parameters list should be a flat list corresponding to the atoms of
    each consecutive structure. All shapes accept a few general parameters, and some
    specific ones

    .. code-block:: python

        # general parameters
        {
            # centering (defaults to origin for structure, atom position for atom)
            "position": [float, float, float],
            # scaling of the size of the shape
            "scale": float,
            # optional, given as quaternion in (x, y, z, w) format
            "orientation": [float, float, float, float],
            "color": string | hex_code,  # e.g. 0xFF0000
        }

        # "kind" : "sphere"
        {
            "radius": float,
        }

        # "kind" : "ellipsoid"
        {
            "semiaxes": [float, float, float],
        }

        # "kind" : "cylinder"
        {
            # "orientation" is redundant and hence ignored
            "vector": [float, float, float],  # orientation and shape of the cylinder
            # the tip of the cylinder is at the end of the segment.
            "radius": float,
        }

        # "kind" : "arrow"
        {
            # "orientation" is redundant and hence ignored
            "vector": [float, float, float],  # orientation and shape of the arrow
            "baseRadius": float,
            "headRadius": float,
            # the tip of the arrow is at the end of the segment.
            # It will extend past the base point if the arrow is not long enough
            "headLength": float,
        }

        # "kind" : "custom"
        {
            "vertices": [  # list of vertices
                [float, float, float],
                ...,
            ],
            # mesh triangulation (optional); computed via convex triangulation
            # where omitted
            "simplices": [
                [int, int, int],  # indices refer to the list of vertices
                ...,
            ],
        }


    .. _`ase.Atoms`: https://wiki.fysik.dtu.dk/ase/ase/atoms.html
    .. _`stk.BuildingBlocks`: https://stk.readthedocs.io/en/stable/_autosummary/stk.BuildingBlock.html#stk.BuildingBlock
    .. _`MDAnalysis.AtomGroup`: https://docs.mdanalysis.org/2.9.0/documentation_pages/core/groups.html#MDAnalysis.core.groups.AtomGroup
    """

    if frames is not None:
        warnings.warn(
            "`frames` argument is deprecated, use `structures` instead",
            stacklevel=2,
        )
        if structures is not None:
            raise ValueError("cannot use both `structures` and `frames` arguments")

        structures = frames

    if meta is not None:
        warnings.warn(
            "`meta` argument is deprecated, use `metadata` instead",
            stacklevel=2,
        )
        if metadata is not None:
            raise ValueError("cannot use both `metadata` and `meta` arguments")

        metadata = meta

    data = {
        "meta": _normalize_metadata(metadata if metadata is not None else {}),
    }

    if settings is not None:
        # dump/load as json to catch possible json incompatibility in settings
        # early
        if not isinstance(settings, dict):
            raise ValueError(
                f"expected 'settings' to be a dict, got {type(settings)} instead"
            )

        data["settings"] = json.loads(json.dumps(settings))

    data["structures"] = []
    n_structures = None

    if structures is not None:
        data["structures"] = structures_to_json(structures)
        n_structures = len(data["structures"])
        n_atoms = sum(s["size"] for s in data["structures"])

    else:
        n_atoms = 0

        # if structures are not given, we create a dataset with only properties.
        # In that case, all properties should be structure properties
        for name, value in properties.items():
            if not isinstance(value, dict):
                if n_structures is None:
                    n_structures = len(value)
                else:
                    if len(value) != n_structures:
                        raise ValueError(
                            f"wrong size for property '{name}': expected "
                            f"{n_structures} elements, but got an array with "
                            f"{len(value)} entries"
                        )
            else:
                if value["target"] != "structure":
                    raise ValueError(
                        f"property '{name}' has a non-structure target, "
                        "which is not allowed if structures are not provided"
                    )

                n_structures = len(value["values"])

    if environments is not None:
        if "structures" not in data:
            raise ValueError("can not have environments without structures")

        data["environments"] = _normalize_environments(environments, data["structures"])
        n_atoms = len(data["environments"])

    if shapes is not None:
        data["shapes"] = _validate_shapes(data["structures"], shapes)

    data["properties"] = {}
    if properties is not None:
        properties = _expand_properties(properties, n_structures, n_atoms)
        for name, value in properties.items():
            multidimensional = False
            if "parameters" in value:
                multidimensional = True
            data["properties"].update(
                _linearize(name, value, n_structures, n_atoms, multidimensional)
            )

    if parameters is not None:
        if not isinstance(parameters, dict):
            raise ValueError(
                f"expecting parameters to be a of type 'dict' not '{type(parameters)}'"
            )

        data["parameters"] = {}
        for key in parameters:
            param = {}
            if not isinstance(parameters[key], dict):
                raise ValueError(
                    f"expecting parameter {key} to be of type 'dict' not "
                    f"'{type(parameters[key])}'"
                )

            if not isinstance(np.asarray(parameters[key]["values"]), np.ndarray):
                raise ValueError(
                    f"expecting the 'values' of parameter {key} to be of type "
                    f"'numpy.ndarray' not '{type(parameters[key]['values'])}'"
                )
            if len(parameters[key]["values"].shape) != 1:
                raise ValueError(
                    f"the 'values' of parameter {key} should be a 1D numpy array"
                )
            param["values"] = _typetransform(
                list(parameters[key]["values"]), f"parameter {key}"
            )

            if "name" in parameters[key]:
                if not isinstance(parameters[key]["name"], str):
                    raise ValueError(
                        f"optional 'name': {parameters[key]['name']} of "
                        f"parameter {key} should a string"
                    )
                param["name"] = parameters[key]["name"]

            if "units" in parameters[key]:
                if not isinstance(parameters[key]["units"], str):
                    raise ValueError(
                        f"optional 'units': {parameters[key]['units']} of "
                        f"parameter {key} should a string"
                    )
                param["units"] = parameters[key]["units"]
            data["parameters"][key] = param

    return data


def write_external_structures(
    structures=None,
    *,
    prefix="structure",
    compresslevel=5,
    # for backward compatibility
    frames=None,
):
    """
    Export a list of structures to external JSON files, and returns a list of
    dictionaries that can be used to create a chemiscope dataset that references them.

    .. code-block:: python

        structures = ase.io.read("trajectory.xyz", ":")

        user_structures = chemiscope.write_external_structures(
            structures, prefix="my_structure"
        )

        write_input("chemiscope.json", structures=user_structures)


    :param list structures: list of atomic structures in a format that can be understood
            by ``chemiscope``.
    :param str prefix: prefix to use for each generated JSON filename. Files will be
        named like ``{prefix}-0.json``, ``{prefix}-1.json`` etc.
    :param str compresslevel: if zero, structures are saved as uncompressed ``.json``
        files, otherwise they are compressed and saved as ``.json.gz`` files, according
        to the desired compression level (1:fast, large file; 9: slow, small file).

    :return: list of paths to JSON file as a list of ``{"size":<natoms>,
        "data":<filename>}`` records that can be used to create a concise chemiscope
        file that references external files as structures
    """
    if frames is not None:
        warnings.warn(
            "`frames` argument is deprecated, use `structures` instead",
            stacklevel=2,
        )
        if structures is not None:
            raise ValueError("cannot use both `structures` and `frames` arguments")

        structures = frames

    json_structures = structures_to_json(structures)

    if "data" in json_structures[0]:
        raise ValueError("`structures` should contain data, got external links instead")

    user_structures = []
    for i, structure in enumerate(json_structures):
        if compresslevel > 0:
            path = f"{prefix}-{i}.json.gz"
            with gzip.open(path, "w", compresslevel) as file:
                file.write(json.dumps(structure, indent=2).encode("utf8"))
        else:
            path = f"{prefix}-{i}.json"
            json.dump(structure, open(path, "w"), indent=2)
        user_structures.append({"size": structure["size"], "data": path})

    return user_structures


def write_input(
    path,
    structures=None,
    *,
    metadata=None,
    properties=None,
    environments=None,
    shapes=None,
    settings=None,
    parameters=None,
    # for backward compatibility
    frames=None,
    meta=None,
):
    """
    Create the input JSON file used by the default chemiscope visualizer, and save it to
    the given ``path``.

    :param str | Path | file-like path: name of the file to use to save the json data.
        If it ends with '.gz', a gzip compressed file will be written

    :param list structures: list of atomic structures.

    :param dict metadata: optional metadata of the dataset

    :param dict properties: optional dictionary of additional properties

    :param list environments: optional list of ``(structure id, atom id, cutoff)``
        specifying which atoms have properties attached and how far out atom-centered
        environments should be drawn by default.

    :param dict shapes: optional dictionary of shapes to have available for display. See
        :py:func:`create_input` for more information on how to define shapes.

    :param dict settings: optional dictionary of settings to use when displaying the
        data. Possible entries for the ``settings`` dictionary are documented in the
        chemiscope input file reference.

    :param dict parameters: optional dictionary of parameters of multidimensional
        properties

    This function uses :py:func:`create_input` to generate the input data, see the
    documentation of this function for more information.

    Here is a quick example of generating a chemiscope input reading the structures from
    a file that `ase <ase-io_>`_ can read, and performing PCA using `sklearn`_ on a
    descriptor computed with another package.

    .. code-block:: python

        import ase
        from ase import io
        import numpy as np

        import sklearn
        from sklearn import decomposition

        import chemiscope

        structures = ase.io.read("trajectory.xyz", ":")

        # example property 1: list containing the energy of each structure,
        # from calculations performed beforehand
        energies = np.loadtxt("energies.txt")

        # example property 2: PCA projection computed using sklearn.
        # X contains a multi-dimensional descriptor of the structure
        X = np.array(...)
        pca = sklearn.decomposition.PCA(n_components=3).fit_transform(X)

        # if the ASE structures also contain additional data, they can be easily
        # extracted as a dictionary using a simple utility function
        structures_properties = chemiscope.extract_properties(
            structures,
            only=["temperature", "classification"],
        )

        # alternatively, properties can also be defined manually
        properties = {
            "PCA": {
                "target": "atom",
                "values": pca,
                "description": "PCA of per-atom representation of the structures",
            },
            "energies": {
                "target": "structure",
                "values": energies,
                "units": "kcal/mol",
            },
        }

        # additional multidimensional properties to be plotted
        dos = np.loadtxt(...)  # load the 2D data
        dos_energy_grid = np.loadtxt(...)
        multidimensional_properties = {
            "DOS": {
                "target": "structure",
                "values": dos,
                "parameters": ["energy"],
            }
        }

        multidimensional_parameters = {
            "energy": {
                "values": dos_energy_grid,
                "units": "eV",
            }
        }

        # merge all properties together
        properties.extend(structures_properties)
        properties.extend(multidimensional_properties)

        chemiscope.write_input(
            path="chemiscope.json.gz",
            structures=structures,
            properties=properties,
            # This is required to display properties with `target: "atom"`
            environments=chemiscope.all_atomic_environments(structures),
            # this is necessary to plot the multidimensional data
            parameters=multidimensional_parameters,
        )

    .. _ase-io: https://wiki.fysik.dtu.dk/ase/ase/io/io.html
    .. _sklearn: https://scikit-learn.org/
    """
    if frames is not None:
        warnings.warn(
            "`frames` argument is deprecated, use `structures` instead",
            stacklevel=2,
        )
        if structures is not None:
            raise ValueError("cannot use both `structures` and `frames` arguments")

        structures = frames

    if meta is not None:
        warnings.warn(
            "`meta` argument is deprecated, use `metadata` instead",
            stacklevel=2,
        )
        if metadata is not None:
            raise ValueError("cannot use both `metadata` and `meta` arguments")

        metadata = meta

    if not (path.endswith(".json") or path.endswith(".json.gz")):
        raise Exception("path should end with .json or .json.gz")

    data = create_input(
        structures=structures,
        metadata=metadata,
        properties=properties,
        environments=environments,
        shapes=shapes,
        settings=settings,
        parameters=parameters,
    )

    if "name" not in data["meta"] or data["meta"]["name"] == "<unknown>":
        data["meta"]["name"] = os.path.basename(path).split(".")[0]

    if isinstance(path, Path):
        path = str(path)

    if isinstance(path, str):
        if path.endswith(".gz"):
            with gzip.open(path, "w", 9) as file:
                file.write(json.dumps(data).encode("utf8"))
        else:
            with open(path, "w") as file:
                json.dump(data, file, indent=2)
    else:
        json.dump(data, path, indent=2)


def _normalize_environments(environments, structures):
    cleaned = []
    for environment in environments:
        if len(environment) != 3:
            raise ValueError(
                f"expected environments to contain three values, got {environment}"
            )

        structure, center, cutoff = environment
        structure = int(structure)
        center = int(center)
        cutoff = float(cutoff)

        if structure >= len(structures):
            raise ValueError(
                f"invalid structure index in environments: got {structure}, "
                f"but we have {len(structures)} structures"
            )

        if center >= structures[structure]["size"]:
            raise ValueError(
                f"invalid center index in environments: got {center} in structure "
                f"which only contains {structures[structure]['size']} atoms"
            )

        if cutoff <= 0:
            raise ValueError("negative cutoff in environments is not valid")

        cleaned.append(
            {
                "structure": structure,
                "center": center,
                "cutoff": float(cutoff),
            }
        )

    return cleaned


def quick_settings(
    *,
    x=None,
    y=None,
    z=None,
    map_color=None,
    structure_color=None,
    size=None,
    symbol=None,
    trajectory=False,
    periodic=False,
    target="structure",
    map_settings=None,
    structure_settings=None,
):
    """A utility function to return a ``settings`` dictionary with the most basic
    options for a chemiscope viewer (e.g. what to show on the axes).

    ``map_settings`` and ``structure_settings`` will override the values if specified,
    e.g. ``quick_settings(x="PCA[1]", map_settings={"x": "energy"})`` will set "energy"
    as x-axis.

    All the parameters are optional.

    :param str x: The property to show on the x axis of the map

    :param str y: The property to show on the y axis of the map.

    :param str z: The property to show on the z axis of the map.

    :param str map_color: The property to use to color the map.

    :param str structure_color: The property to use to color the structure.

    :param str size: The property to use to determine data point size.

    :param str symbol: The (categorical) property to use to determine point markers.

    :param bool trajectory: A boolean flag that sets some default options suitable to
        view trajectory data: fixing the viewpoint for the structure, reducing the delay
        when cycling between structures and adding a line joining the points in the map.

    :param bool periodic: A boolean flag that sets some default options suitable to
        view periodic structures: showing the unit cell, and replicating the structure
        to show atoms at the boundaries of the unit cell.

    :param str target: An option between "atom" and "structure" that determines whether
        the properties shown should be those of environments or of structures.

    :param dict map_settings: Additional settings for the map (following the chemiscope
        settings schema). It will override the settings specied from other map
        parameters, e.g., ``x``, ``y``, ``z`` -axes, ``map_color`` etc.

    :param dict structure_settings: Additional settings for the structure viewer
        (following the chemiscope settings schema). It will override the settings
        specied from other parameters, e.g., ``structure_color``
    """

    if target not in ["atom", "structure"]:
        raise ValueError(
            f"Invalid value {target} for `target`, "
            "should be either `atom` or `structure`."
        )

    if map_settings is not None and not isinstance(map_settings, dict):
        raise TypeError("map_settings must be a dict if specified")
    if structure_settings is not None and not isinstance(structure_settings, dict):
        raise TypeError("structure_settings must be a dict if specified")

    computed_map_settings = {"joinPoints": trajectory}

    properties = {"x": x, "y": y, "z": z, "color": map_color, "size": size}
    for key, value in properties.items():
        if value is not None:
            if not isinstance(value, str):
                raise TypeError(f"{key} must be a string, got {type(value)}")
            computed_map_settings.update({key: {"property": value}})

    if symbol is not None:
        if not isinstance(symbol, str):
            raise TypeError(f"'symbol' must be a string, got {type(symbol)}")
        computed_map_settings.update({"symbol": symbol})

    computed_structure_settings = {
        "keepOrientation": trajectory,
        "playbackDelay": 50 if trajectory else 500,
        "unitCell": periodic,
        "supercell": [3, 3, 3] if periodic and target == "atom" else [1, 1, 1],
    }

    if structure_color is not None:
        if not isinstance(structure_color, str):
            raise TypeError(
                f"{structure_color} must be a string, got {type(structure_color)}"
            )
        computed_structure_settings.update({"color": {"property": structure_color}})

    if map_settings is not None:
        computed_map_settings.update(map_settings)
    if structure_settings is not None:
        computed_structure_settings.update(structure_settings)

    return {
        "map": computed_map_settings,
        "structure": [computed_structure_settings],
        "target": target,
    }


def _normalize_metadata(metadata):
    cleaned = {}
    if "name" in metadata and str(metadata["name"]) != "":
        cleaned["name"] = str(metadata["name"])
    else:
        cleaned["name"] = "<unknown>"

    if "description" in metadata:
        cleaned["description"] = str(metadata["description"])

    if "authors" in metadata:
        cleaned["authors"] = list(map(str, metadata["authors"]))

    if "references" in metadata:
        cleaned["references"] = list(map(str, metadata["references"]))

    for key in metadata.keys():
        if key not in ["name", "description", "authors", "references"]:
            warnings.warn(f"ignoring unexpected metadata: {key}", stacklevel=2)

    return cleaned


def _expand_properties(short_properties, n_structures, n_atoms):
    """
    Convert a shortened entries of properties into the expanded form.
    Entries in already expanded form are not changed.

    :param dict short_properties: properties to handle
    :param int n_structures: number of structures in the dataset
    :param int n_atoms: total number of atoms in the whole dataset

    For example this property dict:
    .. code-block:: python

        properties = {
            "apple": {
                "target": "atom",
                "values": np.zeros((300, 4)),
                "units": "random / fs",
            },
            "orange": np.zeros((100, 42)),
            "banana": np.zeros((300, 17)),
        }

    will be converted to
    .. code-block:: python

        properties = {
            "apple": {
                "target": "atom",
                "values": np.zeros((300, 4)),
                "units": "random / fs",
            },
            "orange": {
                "target": "structure",
                "values": np.zeros((100, 42)),
            },
            "banana": {
                "target": "atom",
                "values": np.zeros((300, 17)),
            },
        }

    assuming that number of structures in the dataset is 100 and
    total number of atoms in the dataset is 300.
    """
    properties = {}
    for key, value in short_properties.items():
        if isinstance(value, dict):
            properties[key] = value
        else:
            is_list = isinstance(value, list)
            is_np_array = isinstance(value, np.ndarray)

            if (not is_list) and (not is_np_array):
                raise ValueError(
                    "Property values should be either list or numpy array, "
                    f"got {type(value)} instead"
                )

            list_has_nan = is_list and any(
                isinstance(v, Number) and np.isnan(v) for v in value
            )
            np_array_has_nan = (
                is_np_array
                and np.issubdtype(value.dtype, np.number)
                and np.isnan(value).any()
            )
            if list_has_nan or np_array_has_nan:
                warnings.warn(
                    f"Property '{key}' has NaNs. If it is selected for display, the "
                    "corresponding points will be hidden from the map",
                    stacklevel=2,
                )

            if n_structures == n_atoms:
                warnings.warn(
                    f"The target of the property '{key}' is ambiguous because "
                    "there is the same number of atoms and structures. "
                    "We will assume target=structure",
                    stacklevel=2,
                )

            dict_property = {"values": value}

            # heuristically determines the type of target
            if len(value) == n_structures:
                dict_property["target"] = "structure"
            elif len(value) == n_atoms:
                dict_property["target"] = "atom"
            else:
                raise ValueError(
                    "The length of property values is different from the "
                    "number of structures and the number of atoms, we can not "
                    f"guess the target. Got n_atoms = {n_atoms}, n_structures = "
                    f"{n_structures}, the length of property values is "
                    f"{len(value)}, for the '{key}' property"
                )

            properties[key] = dict_property
    return properties


def _linearize(name, property, n_structures, n_centers, multidimensional=False):
    """
    Transform a single property dict (containing "value", "target", "units",
    "description", "parameters") with potential multi-dimensional "values" key
    to data that chemiscope can load.

    Multi-dimensional "value" generate multiple properties named "XXX [1]", "XXX
    [2]", "XXX [3]", etc. Data in "values" are converted to either string or
    float, to ensure it is compatible with JSON.

    :param name: name of the property related to this data, to be used in error
                 messages
    :param property: dictionary containing the property data and metadata
    :param n_structures: total number of structures, to validate the array sizes
    :param n_centers: total number of atoms, to validate the array sizes
    :param param: describes if the property is multisimensional to be plotted
    """
    _validate_property(name, property)

    data = {}
    values = property["values"]
    if not multidimensional:
        if isinstance(values, list) and (
            isinstance(values[0], list) or isinstance(values[0], np.ndarray)
        ):
            # convert to ndarray so we can use the parser below
            values = np.array(values)
        if isinstance(values, list):
            # deal with simple list
            data[name] = {
                "target": property["target"],
                "values": _typetransform(values, name),
            }
        elif isinstance(values, np.ndarray):
            if len(values.shape) == 1:
                data[name] = {
                    "target": property["target"],
                    "values": _typetransform(list(values), name),
                }
            elif len(values.shape) == 2:
                if values.shape[1] == 1:
                    data[name] = {
                        "target": property["target"],
                        "values": _typetransform(list(values), name),
                    }
                else:
                    for i in range(values.shape[1]):
                        data[f"{name}[{i + 1}]"] = {
                            "target": property["target"],
                            "values": _typetransform(list(values[:, i]), name),
                        }
            else:
                raise Exception("unsupported ndarray property")
        else:
            raise Exception(
                f"unknown type ({type(property['values'])}) for property '{name}'"
            )
    else:
        assert isinstance(property["parameters"][0], str)

        ndvalues = []
        for i in range(len(values)):
            ndvalues += [_typetransform(list(values[i]), name)]
        data[name] = {
            "target": property["target"],
            "values": ndvalues,
            "parameters": property["parameters"],
        }

    # get property metadata
    if "units" in property:
        for item in data.values():
            item["units"] = str(property["units"])

    if "description" in property:
        for i, item in enumerate(data.values()):
            # add [component XX] to the description if values was a ndarray
            extra = f" [component {i + 1}]" if len(data) > 1 else ""
            item["description"] = str(property["description"]) + extra

    # Validate the properties size
    for prop in data.values():
        if prop["target"] == "atom" and len(prop["values"]) != n_centers:
            raise Exception(
                f"wrong size for the property '{name}' with target=='atom': "
                f"expected {n_centers} values, got {len(prop['values'])}"
            )

        if prop["target"] == "structure" and len(prop["values"]) != n_structures:
            raise Exception(
                f"wrong size for the property '{name}' with target=='structure': "
                f"expected {n_structures} values, got {len(prop['values'])}"
            )

    return data


def _validate_property(name, property):
    if name == "":
        raise Exception("the name of a property can not be the empty string")
    elif not isinstance(name, str):
        raise Exception(
            "the name of a property name must be a string, "
            f"got '{name}' of type {type(name)}"
        )

    if "target" not in property:
        raise Exception(f"missing 'target' for the '{name}' property")
    elif property["target"] not in ["atom", "structure"]:
        raise Exception(
            f"the target must be 'atom' or 'structure' for the '{name}' property"
        )

    if "values" not in property:
        raise Exception(f"missing 'values' for the '{name}' property")

    if "parameters" in property:
        if len(property["parameters"]) != 1 or not isinstance(
            property["parameters"][0], str
        ):
            raise Exception(
                f"the parameter of property {name} should be a list containing "
                "a single string"
            )

    for key in property.keys():
        if key not in ["target", "values", "description", "units", "parameters"]:
            warnings.warn(f"ignoring unexpected property key: {key}", stacklevel=2)


def _typetransform_scalar(data):
    """Transform a single scalar value to either string or float."""
    if isinstance(data, (str, np.str_)):
        return str(data)
    elif isinstance(data, bytes):
        return data.decode("utf8")
    else:
        return float(data)


def _typetransform(data, name):
    """
    Transform the given data to either a list of string or a list of floats.

    :param data: list of unknown type to be converted
    :param name: name of the property related to this data, to be used in
                 error messages
    """
    assert isinstance(data, list) and len(data) > 0

    try:
        values = []
        for value in data:
            if isinstance(value, np.ndarray) and value.shape == (1,):
                values.append(_typetransform_scalar(value[0]))
            else:
                values.append(_typetransform_scalar(value))

        return values
    except Exception:
        raise Exception(
            f"unsupported type in property '{name}' values: should be string or number"
        )


def _validate_shapes(structures, shapes):
    if not isinstance(shapes, dict):
        raise TypeError(f"`shapes` must be a dictionary, got {type(shapes)} instead")

    # validate type and number of element for each entries in the shapes
    for key, shapes_for_key in shapes.items():
        if not isinstance(key, str):
            raise TypeError(
                f"the `shapes` dictionary keys must be strings, got {type(key)}"
            )

        if not isinstance(shapes_for_key, dict):
            raise TypeError(
                "Each entry in `shapes` must be a dictionary, "
                f"got {type(shapes_for_key)} instead for '{key}'"
            )

        for shape_key in shapes_for_key:
            if shape_key not in ["kind", "parameters"]:
                raise ValueError(
                    f"Invalid entry `{shape_key}` in the specification "
                    f"for shape `{key}`"
                )

        base_shape = shapes_for_key["parameters"].get("global", {})
        structure_parameters = shapes_for_key["parameters"].get("structure", None)
        atom_parameters = shapes_for_key["parameters"].get("atom", None)
        atom_counter = 0

        for structure_i in range(len(structures)):
            if (
                structure_parameters is not None
                and len(structure_parameters) <= structure_i
            ):
                raise TypeError(
                    f"structure_parameters must be a list with {(len(structures))} "
                    f"elements, got {len(structure_parameters)}"
                )
            for _ in range(structures[structure_i]["size"]):
                if atom_parameters is not None and len(atom_parameters) <= atom_counter:
                    raise TypeError(
                        f"atom_parameters must be a list coinciding to the atomic "
                        f"environments, got {len(atom_parameters)} elements"
                    )

                shape = {
                    "kind": shapes_for_key["kind"],
                    "parameters": {
                        "global": base_shape,
                        "structure": (
                            structure_parameters[structure_i]
                            if structure_parameters is not None
                            else {}
                        ),
                        "atom": (
                            atom_parameters[atom_counter]
                            if atom_parameters is not None
                            else {}
                        ),
                    },
                }
                _check_valid_shape(shape)
                atom_counter += 1

    for shape in shapes.values():
        if (
            shape["kind"] == "custom"
            and "vertices" in shape["parameters"]
            and "simplices" not in shape["parameters"]
        ):
            try:
                import scipy.spatial
            except ImportError as e:
                raise RuntimeError(
                    "Missing simplices in custom shape, and scipy is not installed"
                ) from e

            convex_hull = scipy.spatial.ConvexHull(shape["parameters"]["vertices"])
            shape["parameters"]["simplices"] = [
                s.tolist() for s in convex_hull.simplices
            ]
    return shapes


def _check_valid_shape(shape):
    if not isinstance(shape, dict):
        raise TypeError(
            f"individual shapes must be dictionaries, got {type(shape)} instead"
        )

    always_okay = ["orientation", "scale", "position", "color"]
    parameters = {}
    if "parameters" in shape:
        parameters.update(shape["parameters"]["global"])
    if "structure" in shape["parameters"]:
        parameters.update(shape["parameters"]["structure"])
    if "atom" in shape["parameters"]:
        parameters.update(shape["parameters"]["atom"])

    if len(parameters) == 0:
        raise ValueError(f"no parameters provided for {shape['kind']} shape")
    if shape["kind"] == "sphere":
        for parameter in parameters:
            if parameter not in ["radius", *always_okay]:
                raise ValueError(
                    f"unknown shape parameter '{parameter}' for 'sphere' shape kind"
                )

        if not isinstance(parameters["radius"], float):
            raise TypeError(
                "sphere shape 'radius' must be a float, got "
                f"{type(parameters['radius'])}"
            )

    elif shape["kind"] == "ellipsoid":
        for parameter in parameters.keys():
            if parameter not in ["semiaxes", *always_okay]:
                raise ValueError(
                    f"unknown shape parameter '{parameter}' for 'ellipsoid' shape kind"
                )

        semiaxes_array = np.asarray(parameters["semiaxes"]).astype(
            np.float64, casting="safe", subok=False, copy=False
        )

        if not semiaxes_array.shape == (3,):
            raise ValueError(
                "'semiaxes' must be an array with 3 values for 'ellipsoid' shape kind"
            )

    elif shape["kind"] == "custom":
        for parameter in parameters.keys():
            if parameter not in ["vertices", "simplices", *always_okay]:
                raise ValueError(
                    f"unknown shape parameter '{parameter}' for 'custom' shape kind"
                )

        vertices_array = np.asarray(parameters["vertices"]).astype(
            np.float64, casting="safe", subok=False, copy=False
        )

        if len(vertices_array.shape) != 2 or vertices_array.shape[1] != 3:
            raise ValueError(
                "'vertices' must be an Nx3 array values for 'custom' shape kind"
            )

        if "simplices" in parameter:
            simplices_array = np.asarray(parameters["simplices"]).astype(
                int, casting="safe", subok=False, copy=False
            )

            if len(simplices_array.shape) != 2 or simplices_array.shape[1] != 3:
                raise ValueError(
                    "'simplices' must be an Nx3 array values for 'custom' shape kind"
                )
    elif shape["kind"] == "cylinder":
        if not isinstance(parameters["radius"], float):
            raise TypeError(
                "cylinder shape 'radius' must be a float, "
                f"got {type(parameters['radius'])}"
            )
        vector_array = np.asarray(parameters["vector"]).astype(
            np.float64, casting="safe", subok=False, copy=False
        )

        if not vector_array.shape == (3,):
            raise ValueError(
                "'vector' must be an array with 3 values for 'cylinder' shape kind"
            )
    elif shape["kind"] == "arrow":
        if not isinstance(parameters["baseRadius"], float):
            raise TypeError(
                "arrow shape 'baseRadius' must be a float, "
                f"got {type(parameters['baseRadius'])}"
            )
        if not isinstance(parameters["headRadius"], float):
            raise TypeError(
                "arrow shape 'headRadius' must be a float, "
                f"got {type(parameters['headRadius'])}"
            )
        if not isinstance(parameters["headLength"], float):
            raise TypeError(
                "arrow shape 'headLength' must be a float, "
                f"got {type(parameters['headLength'])}"
            )

        vector_array = np.asarray(parameters["vector"]).astype(
            np.float64, casting="safe", subok=False, copy=False
        )

        if not vector_array.shape == (3,):
            raise ValueError(
                "'vector' must be an array with 3 values for 'arrow' shape kind"
            )
    else:
        raise ValueError(f"unknown shape kind '{shape['kind']}'")

    if "orientation" in parameters:
        orientation_array = np.asarray(parameters["orientation"]).astype(
            np.float64, casting="safe", subok=False, copy=False
        )

        if not orientation_array.shape == (4,):
            raise ValueError(
                "semiaxes must be an array with 4 values for 'ellipsoid' shape kind"
            )



import warnings
from pathlib import Path


mta = None
mts = None
torch = None
vesin_metatomic = None


class metatomic_featurizer:
    """
    Create a featurizer function using a `metatomic`_ model to obtain the features from
    structures. The model must be able to create a ``"features"`` output.

    :param model: model to use for the calculation. It can be a file path, a Python
        instance of :py:class:`metatomic.torch.AtomisticModel`, or the output of
        :py:func:`torch.jit.script` on :py:class:`metatomic.torch.AtomisticModel`.
    :param extensions_directory: a directory where model extensions are located
    :param check_consistency: should we check the model for consistency when running,
        defaults to False.
    :param device: a torch device to use for the calculation. If ``None``, the function
        will use the options in model's ``supported_device`` attribute.
    :param length_unit: Unit of length used in the structures.
    :param variant: selects which feature output variant to use. By default, the main
        ``"features"`` output is used. To choose another variant, provide its name
        (e.g., ``"cos_sin"``), which will select the corresponding
        ``"features/<variant>"`` output.

    :returns: a function that takes a list of structures and returns the features.

    To use this function, additional dependencies are required. They can be installed
    with the following command:

    .. code:: bash

        pip install chemiscope[explore]

    Here is an example using a pre-trained `metatomic`_ model, stored as a ``model.pt``
    file with the compiled extensions stored in the ``extensions/`` directory. The
    structures are obtained by reading structures from a file that `ase <ase-io_>`_ can
    read.

    .. code-block:: python

        import chemiscope
        import ase.io

        # Read the structures from the dataset
        structures = ase.io.read("data/explore_c-gap-20u.xyz", ":")

        # Provide model file ("model.pt") to `metatensor_featurizer`
        featurizer = chemiscope.metatensor_featurizer(
            "model.pt", extensions_directory="extensions"
        )

        chemiscope.explore(structures, featurizer=featurizer)

    For more examples, see the related :ref:`documentation
    <chemiscope-explore-metatomic>`.

    .. _metatomic: https://docs.metatensor.org/metatomic/
    """

    def __init__(
        self,
        model,
        *,
        extensions_directory=None,
        check_consistency=None,
        device=None,
        length_unit="Angstrom",
        variant=None,
    ):
        # Check if dependencies were installed
        global mta, mts, torch, vesin_metatomic
        if mta is None or mts is None or torch is None or vesin_metatomic is None:
            try:
                import metatensor.torch as mts
                import metatomic.torch as mta
                import torch
                import vesin.metatomic as vesin_metatomic
            except ImportError as e:
                raise ImportError(
                    f"Required package not found: {e}. Please install the "
                    "dependencies with `pip install chemiscope[explore]`."
                )

        if isinstance(model, (str, Path)):
            self.model = mta.load_atomistic_model(
                model, extensions_directory=extensions_directory
            )
        else:
            self.model = model

        self.length_unit = length_unit
        self.check_consistency = check_consistency

        capabilities = self.model.capabilities()

        self.feature_output_name = mta.pick_output(
            "features", capabilities.outputs, variant
        )

        if capabilities.dtype == "float32":
            self.dtype = torch.float32
        else:
            assert capabilities.dtype == "float64"
            self.dtype = torch.float64

        self.device = _find_best_device(capabilities.supported_devices, device)

    def __call__(self, structures, environments):
        systems = mta.systems_to_torch(structures)
        vesin_metatomic.compute_requested_neighbors(
            systems,
            self.length_unit,
            self.model,
        )

        systems = [s.to(self.dtype, self.device) for s in systems]

        if environments is not None:
            capabilities = self.model.capabilities()
            if not capabilities.outputs[self.feature_output_name].per_atom:
                raise ValueError(
                    (
                        "this model does not support per-atom features calculation for "
                        f"'{self.feature_output_name}' output"
                    )
                )

            selected_atoms = mts.Labels(
                names=["system", "atom"],
                values=torch.tensor(
                    [(system, atom) for system, atom, _ in environments]
                ),
            )
        else:
            selected_atoms = None

        options = mta.ModelEvaluationOptions(
            length_unit=self.length_unit,
            outputs={
                self.feature_output_name: mta.ModelOutput(
                    per_atom=environments is not None
                )
            },
            selected_atoms=selected_atoms,
        )

        outputs = self.model(
            systems,
            options,
            check_consistency=self.check_consistency,
        )

        # outputs should already be in the correct order because we passed
        # `selected_atoms` to the model
        return outputs[self.feature_output_name].block().values.detach().cpu().numpy()


def _find_best_device(supported_devices, requested_device):
    """
    Find the best device from the list of ``devices`` that is available to the current
    PyTorch installation.
    """

    available = []
    for device in supported_devices:
        if device == "cpu":
            available.append("cpu")
        elif device == "cuda":
            if torch.cuda.is_available():
                available.append("cuda")
        elif device == "mps":
            if (
                hasattr(torch.backends, "mps")
                and torch.backends.mps.is_built()
                and torch.backends.mps.is_available()
            ):
                available.append("mps")
        else:
            warnings.warn(
                f"unknown device in the model's `supported_devices`: '{device}'",
                stacklevel=3,
            )

    if requested_device is None:
        if len(available) == 0:
            warnings.warn(
                "could not find a valid device in the model's `supported_devices`, "
                "falling back to CPU",
                stacklevel=3,
            )
            return "cpu"

        return available[0]

    if requested_device in available:
        return requested_device

    warnings.warn(
        f"the requested device '{requested_device}' is not available on this machine "
        "or not supported by this model, falling back to CPU",
        stacklevel=3,
    )
    return "cpu"



import warnings

import numpy as np


try:
    import MDAnalysis as mda

    HAVE_MDA = True
except ImportError:
    HAVE_MDA = False

BIO_PROPERTIES = ["hetatom", "resids", "resnames", "chains"]


def _mda_valid_structures(structures):
    if HAVE_MDA and isinstance(structures, mda.AtomGroup):
        return structures, True
    else:
        return [], False


def _mda_to_json(ag):
    data = {}
    data["size"] = len(ag)
    data["elements"] = (
        ag.elements.tolist() if hasattr(ag, "elements") else ag.types.tolist()
    )
    # `element` is better, but not always available, e.g. xyz file
    if hasattr(ag, "names"):
        data["names"] = ag.names.tolist()
    else:
        data["names"] = data["elements"]
    x, y, z = ag.positions.T
    data["x"] = x.tolist()
    data["y"] = y.tolist()
    data["z"] = z.tolist()
    hetatom = np.full(ag.n_atoms, True)

    if ag.dimensions is not None:
        data["cell"] = list(
            np.concatenate(
                mda.lib.mdamath.triclinic_vectors(ag.dimensions),
                dtype=np.float64,
                # should be np.float64 otherwise not serializable
            )
        )
    if hasattr(ag, "chainIDs") and ag.chainIDs is not None:
        data["chains"] = ag.chainIDs.tolist()
    elif hasattr(ag, "segids") and ag.segids is not None:
        # segids are sometimes abused to store chain ids in PDBs, so we use them here
        data["chains"] = ag.segids.tolist()

    if hasattr(ag, "resnames") and ag.resnames is not None:
        data["resnames"] = [
            resname if resname is not None and resname != "" else "UNK"
            for resname in ag.resnames
        ]
        # atom selection requires the `resname`
        hetatom[ag.select_atoms("protein or nucleic").indices] = False

    if hasattr(ag, "resids") and ag.resids is not None:
        data["resids"] = ag.resids.view(dtype=int).tolist()

    if hasattr(ag, "bonds") and ag.bonds is not None:
        data["bonds"] = np.hstack(
            (ag.bonds.indices, np.full((len(ag.bonds), 1), 1))
        ).tolist()

    data["hetatom"] = hetatom.tolist()

    # remove bio-related properties if any of them are missing
    existing_properties = []
    for prop in BIO_PROPERTIES:
        if prop in data:
            existing_properties.append(prop)

    if len(existing_properties) != len(BIO_PROPERTIES):
        for prop in existing_properties:
            del data[prop]

    return data


def _mda_get_atom_properties(ag) -> dict:
    POSSIBLE_PROPERTIES = ["velocities", "forces"]
    all_properties = {}
    extra = set()
    for property in POSSIBLE_PROPERTIES:
        try:
            all_properties[property] = []
            all_properties[property].append([getattr(ag, property)])
        except mda.exceptions.NoDataError:
            del all_properties[property]

    if len(ag.universe.trajectory) > 1:
        for _ in ag.universe.trajectory[1:]:
            for property in all_properties:
                try:
                    all_properties[property].append([getattr(ag, property)])
                except mda.exceptions.NoDataError:
                    extra.add(property)

    if len(extra) != 0:
        warnings.warn(
            "the following per-atom properties are only defined for a subset "
            f"of structures: {list(sorted(extra))}; they will be ignored",
            stacklevel=2,
        )

    return all_properties


def _mda_atom_properties(ag, only=None, atoms_mask=None):
    all_properties = _mda_get_atom_properties(ag)
    if only is None:
        selected = all_properties
    else:
        selected = {}
        for name in only:
            if name in all_properties.keys():
                selected[name] = all_properties[name]

    # create property in the format expected by create_input
    properties = {
        name: {"target": "atom", "values": np.concatenate(value)}
        for name, value in selected.items()
    }

    if atoms_mask is not None:
        # only include values for requested atoms
        for property in properties.values():
            property["values"] = property["values"][atoms_mask]

    return properties


def _mda_extract_properties(ag, only=None, environments=None):
    """implementation of ``extract_properties`` for MDAnalysis"""
    properties = {}

    if environments is not None:
        atoms_mask = [[False] * len(f) for f in ag.universe.trajectory]
        for structure, center, _ in environments:
            atoms_mask[structure][center] = True

        atoms_mask = np.concatenate(atoms_mask)
    else:
        atoms_mask = None

    atom_properties = _mda_atom_properties(ag, only, atoms_mask)

    for name, values in atom_properties.items():
        if name in properties:
            warnings.warn(
                f"a property named '{name}' is defined for both atoms and structures, "
                "the atom one will be ignored",
                stacklevel=2,
            )
        else:
            properties[name] = values

    return properties


def _mda_all_atomic_environments(ag, cutoff):
    "Extract all atomic environments out of a set of MDAnalysis AtomGroup objects"
    environments = []
    for structure_i, structure in enumerate(ag.universe.trajectory):
        for atom_i in range(len(structure)):
            environments.append((structure_i, atom_i, cutoff))
    return environments



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json

from .input import write_input
from .structures import all_atomic_environments, extract_properties


def _chemiscope_input_parser():
    import argparse

    # Tweak the autogenerated help output to look nicer
    def formatter(prog):
        return argparse.HelpFormatter(prog, max_help_position=22)

    parser = argparse.ArgumentParser(
        description=main.__doc__, formatter_class=formatter
    )
    parser.add_argument(
        "input", type=str, help="input file containing the structures and properties"
    )
    parser.add_argument(
        "-o", "--output", type=str, help="chemiscope output file in JSON format"
    )
    parser.add_argument(
        "--properties",
        type=str,
        default="",
        help=(
            "comma-separated list of properties that should be extracted. "
            "defaults to all"
        ),
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--only-atoms",
        action="store_true",
        help="only use per-atom properties from the input file",
    )
    group.add_argument(
        "--only-structures",
        action="store_true",
        help="only use per-structure properties from the input file (default)",
    )
    parser.add_argument(
        "--cutoff",
        type=float,
        default=3.5,
        help="spherical cutoff radius that should be visualized around environments",
    )
    parser.add_argument("--name", default="", type=str, help="name of the dataset")
    parser.add_argument(
        "--description", default="", type=str, help="description of the dataset"
    )
    parser.add_argument(
        "--authors", nargs="*", type=str, default=[], help="list of dataset authors"
    )
    parser.add_argument(
        "--references",
        nargs="*",
        type=str,
        default=[],
        help="list of references for the dataset",
    )
    parser.add_argument(
        "--settings",
        type=str,
        default="",
        help=(
            "visualization settings, as a JSON string, following the chemiscope format"
        ),
    )
    return parser


def main():
    """
    Command-line utility to generate an input for chemiscope — the interactive
    structure-property explorer. Parses an input file containing atomic structures using
    the ASE I/O module, and converts it into a JSON file that can be loaded in
    chemiscope. Structure and environment properties must be written in the same file
    containing atomic structures: we recommend the extended xyz format, which is
    flexible and simple. In all cases, this utility will simply write to the JSON file
    anything that is readable by ASE.
    """

    try:
        # command-line execution. requires ASE IO module
        import ase.io as ase_io
    except ImportError:
        raise ImportError(
            "chemiscope_input needs ASE modules to parse structure inputs"
        )

    parser = _chemiscope_input_parser()
    args = parser.parse_args()

    if args.only_atoms and args.cutoff is None:
        raise Exception("--only-atoms requires to give --cutoff")
    if args.only_structures and args.cutoff is not None:
        raise Exception("--only-structure can not be given with --cutoff")

    # read file with ASE and remove extraneous properties
    structures = ase_io.read(args.input, ":")
    if args.only_structures:
        for structure in structures:
            for key in list(structure.arrays.keys()):
                if key not in ["positions", "numbers", "momenta"]:
                    del structure.arrays[key]
        environments = None
    elif args.only_atoms:
        for structure in structures:
            structure.info = {}
        environments = all_atomic_environments(structures, cutoff=args.cutoff)

    if args.properties == "":
        filter_properties = None
    else:
        filter_properties = args.properties.split(",")

    # determine output file name automatically if missing
    output = args.output or args.input + "_chemiscope.json.gz"
    properties = extract_properties(structures, only=filter_properties)
    has_environment_property = False
    for prop in properties.values():
        if prop["target"] == "atom":
            has_environment_property = True
            break
    if has_environment_property:
        # assumes properties are associated with all atomic environments
        environments = all_atomic_environments(structures, cutoff=args.cutoff)
    else:
        environments = None

    if args.settings == "":
        settings = None
    else:
        # reads settings as a dictionary, following chemiscope format and schema
        settings = json.loads(args.settings)

    write_input(
        path=output,
        structures=structures,
        properties=properties,
        environments=environments,
        metadata={
            "name": args.name,
            "description": args.description,
            "authors": args.authors,
            "references": args.references,
        },
        settings=settings,
    )


if __name__ == "__main__":
    main()



import os
import re
import uuid

from docutils import nodes


class chemiscope(nodes.Element):
    pass


def visit_chemiscope_latex(self, node):
    self.body.append(
        "\\\\(The chemiscope visualisation cannot be rendered to LaTeX)\\\\\n"
    )


def depart_chemiscope_latex(self, node):
    pass


def visit_chemiscope_html(self, node):
    # determines the path of the static resources
    # relative to the output document
    html_file_path = self.builder.get_outfilename(self.builder.current_docname)
    html_file_dir = os.path.dirname(html_file_path)
    # _static is the hardcoded output folder for static files
    static_dir = os.path.join(self.builder.outdir, "_static")
    static_path = os.path.relpath(static_dir, html_file_dir)

    self.body.append(
        generate_html_content(
            node["filepath"],
            node["mode"],
            node["warning_timeout"],
            node["include_headers"],
            static_path,
        )
    )


def depart_chemiscope_html(self, node):
    pass


def generate_html_content(
    filepath,
    mode="default",
    warning_timeout=4000,
    include_headers=True,
    static_path="_static",
):
    """
    Generate HTML content for displaying a chemiscope visualization.

    Parameters:
    - filepath (str): The path to the generated dataset
    - mode (str): The display mode for the visualization
    - warning_timeout (float): The timeout for hiding warning messages (ms)
        Set to -1 to disable warnings, and to zero to make them persistent.

    Returns:
    - str: The chemiscope widget wrapped in the html
    """
    # Generate a unique id for the chemiscope div
    div_id = f"chsp-{uuid.uuid4()}"

    # Read the template html file
    current_folder_path = os.path.dirname(__file__)
    template_path = os.path.join(
        current_folder_path, "static", "chemiscope-sphinx.html"
    )
    with open(template_path, "r") as file:
        html_template = file.read()

    # Choose whether to discard headers
    if not include_headers:
        html_template = re.sub(
            r"<!-- begin headers -->.*?<!-- end headers -->",
            "",
            html_template,
            flags=re.DOTALL,
        )

    # Replace html placeholders  with actual values
    return (
        html_template.replace("{{div_id}}", div_id)
        .replace("{{filepath}}", filepath)
        .replace("{{mode}}", mode)
        .replace("{{timeout}}", str(warning_timeout))
        .replace("{{static_path}}", static_path)
    )



import os

from ._directive import ChemiscopeDirective
from ._nodes import (
    chemiscope,
    depart_chemiscope_html,
    depart_chemiscope_latex,
    visit_chemiscope_html,
    visit_chemiscope_latex,
)
from ._scraper import ChemiscopeScraper  # noqa: F401


def setup(app):
    app.add_directive("chemiscope", ChemiscopeDirective)
    app.add_node(
        chemiscope,
        override=True,
        html=(visit_chemiscope_html, depart_chemiscope_html),
        latex=(visit_chemiscope_latex, depart_chemiscope_latex),
    )

    def add_static_path(app):
        static_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "static"))
        # put our path first, so the user can override our files if they want to
        app.config.html_static_path.insert(0, static_path)

    app.connect("builder-inited", add_static_path)



from typing import Optional

import numpy as np


def arrow_from_vector(
    position,
    *,
    scale: Optional[float] = 1.0,
    radius: Optional[float] = 0.1,
    head_radius_scale: Optional[float] = 1.75,
    head_length_scale: Optional[float] = 2.0,
):
    """
    Draws an arrow from the origin to the specified 3D ``position``. Returns a custom
    shape in the form required by the chemiscope input. Use ``None`` for the arrow shape
    parameters to leave them undefined (so that they can be specified in the global
    parameters).

    :param scale: conversion from the units of the vector to the units of the atomic
        positions (usually Å)
    :param radius: radius of the stem of the arrow (same units as the atomic positions,
        typically Å)
    :param head_radius_scale: radius of the arrow tip, relative to the stem radius
    :param head_length_scale: length of the arrow tip, relative to the stem radius
    """

    data = {"vector": [v * scale for v in position]}
    if radius is not None:
        data["baseRadius"] = radius
    if head_radius_scale is not None:
        data["headRadius"] = radius * head_radius_scale
    if head_length_scale is not None:
        data["headLength"] = radius * head_length_scale

    return data


def ellipsoid_from_tensor(
    tensor,
    *,
    scale=1.0,
    force_positive=False,
):
    """
    Returns an ellipsoid (semiaxes + quaternion) representation of a positive definite
    tensor (e.g. a polarizability), in the form required by the chemiscope input.

    :param tensor: a positive-definite tensor (3x3 or a 6-array [xx,yy,zz,xy,xz,yz])
    :param scale: conversion from the units of the tensor to the units of the atomic
        positions (usually Å)
    :param force_positive: takes the absolute value of eigenvalues, to handle
        non-positive tensors
    """

    tensor = np.asarray(tensor) * (scale**2)
    if len(tensor.flatten()) == 9:
        matrix = tensor.reshape((3, 3))
    elif len(tensor) == 6:
        matrix = np.array(
            [
                [tensor[0], tensor[3], tensor[4]],
                [tensor[3], tensor[1], tensor[5]],
                [tensor[4], tensor[5], tensor[2]],
            ]
        )
    # Compute the eigenvalues and eigenvectors
    eigenvalues, eigenvectors = np.linalg.eigh(matrix)

    if force_positive:
        eigenvalues = np.abs(eigenvalues)

    # The lengths of the principal axes are the square roots of the eigenvalues
    old_settings = np.seterr(invalid="raise")
    try:
        ax = np.sqrt(eigenvalues[0])
        ay = np.sqrt(eigenvalues[1])
        az = np.sqrt(eigenvalues[2])
    except FloatingPointError as e:
        raise ValueError(
            f"Non-positive definite tensor found with eigenvalues {eigenvalues}.\n"
            "If this is acceptable, set `force_positive=True` to take the "
            "absolute values.",
        ) from e
    np.seterr(**old_settings)

    # makes sure the rotation is proper
    if np.linalg.det(eigenvectors) < 0:
        eigenvectors[:, 0] *= -1

    # Form the rotation matrix from eigenvectors
    rotation = eigenvectors.T

    # converts to quaternion
    try:
        from scipy.spatial.transform import Rotation
    except ImportError as e:
        raise RuntimeError(
            "scipy is required to construct ellipsoids from tensors"
        ) from e
    quaternion = Rotation.from_matrix(rotation.T).as_quat()

    return dict(
        semiaxes=[ax, ay, az],
        orientation=list(quaternion),
    )



#!/usr/bin/env python3
import glob
import json
import os
import shutil
import subprocess
import sys
import warnings

from setuptools import setup
from setuptools.command.bdist_egg import bdist_egg


class bdist_egg_disabled(bdist_egg):
    """Disabled version of bdist_egg

    Prevents setup.py install performing setuptools' default easy_install,
    which it should never ever do.
    """

    def run(self):
        sys.exit(
            "Aborting implicit building of eggs. "
            + "Use `pip install .` or `python setup.py bdist_wheel && pip "
            + "uninstall chemiscope -y && pip install dist/chemiscope-*.whl` "
            + "to install from source."
        )


NPM_BUILD_INPUT = [
    # source files
    *glob.glob("src/**/*", recursive=True),
    *glob.glob("python/jupyter/src/**/*", recursive=True),
    # dependencies
    "package.json",
    "package-lock.json",
    # build configuration
    "tsconfig.json",
    "webpack.config.ts",
    # streamlit component
    *glob.glob("python/streamlit/src/**/*", recursive=True),
]

NPM_BUILD_OUTPUT = [
    # jupyter extensions
    "python/jupyter/nbextension/chemiscope.min.js",
    "python/jupyter/labextension/package.json",
    # sphinx extension
    "python/chemiscope/sphinx/static/chemiscope.min.js",
    # streamlit component
    "python/chemiscope/streamlit/index.html",
    "python/chemiscope/streamlit/chemiscope-streamlit.min.js",
]


def needs_npm_build():
    """
    Determine if we need to re-build the JS code with npm by checking the modification
    time for inputs & outputs of this build.
    """
    last_output_modification_time = -1
    for file in NPM_BUILD_OUTPUT:
        if os.path.exists(file):
            last_output_modification_time = max(
                os.path.getmtime(file), last_output_modification_time
            )
        else:
            # if the file does not exist, we need to build
            return True

    last_input_modification_time = -1
    for file in NPM_BUILD_INPUT:
        if os.path.exists(file):
            last_input_modification_time = max(
                os.path.getmtime(file), last_input_modification_time
            )
        # if the file does not exist, we might be building a sdist which already
        # contains the pre-built extensions

    return last_output_modification_time < last_input_modification_time


def run_npm_build():
    """Build the JavaScript code required by the jupyter widget using npm"""
    root = os.path.dirname(os.path.realpath(__file__))

    # This file does not exist when building a wheel (for pip installation)
    # from a sdist. This is fine since the sdist already contains the required
    # js files.
    if os.path.exists(os.path.join(root, "package.json")):
        subprocess.run("npm ci", check=True, shell=True)
        # build the labextension first, since it triggers a rebuild of the main
        # package code
        try:
            import jupyterlab  # noqa

            subprocess.run("npm run build:labextension", check=True, shell=True)
            with open(
                os.path.join(root, "python", "jupyter", "labextension", "install.json"),
                "w",
            ) as fd:
                install_json = {
                    "__comment": "metadata for chemiscope jupyterlab extension",
                    "packageManager": "python",
                    "packageName": "chemiscope",
                    "uninstallInstructions": "Use your Python package manager "
                    + "(pip, conda, etc.) to uninstall chemiscope",
                }
                json.dump(install_json, fd)
        except ImportError:
            # skip building jupyterlab if it is not installed on the developer
            # machine
            warnings.warn(
                "skipping the lab extension build since jupyterlab is not installed",
                stacklevel=1,
            )

            # still build the main code for the sphinx extension
            subprocess.run("npm run build", check=True, shell=True)

        shutil.copyfile(
            src="dist/chemiscope.min.js",
            dst="python/chemiscope/sphinx/static/chemiscope.min.js",
        )

        subprocess.run("npm run build:nbextension", check=True, shell=True)

        subprocess.run("npm run build:streamlit", check=True, shell=True)

        streamlit_output_dir = os.path.join("python", "chemiscope", "streamlit")
        os.makedirs(streamlit_output_dir, exist_ok=True)
        shutil.copyfile(
            os.path.join("python", "streamlit", "build", "chemiscope-streamlit.min.js"),
            os.path.join(streamlit_output_dir, "chemiscope-streamlit.min.js"),
        )

        shutil.copyfile(
            os.path.join("python", "streamlit", "src", "index.html"),
            os.path.join(streamlit_output_dir, "index.html"),
        )


if __name__ == "__main__":
    # we need to run npm build outside of the call to setup to be able to get
    # the full list of files to install for the labextension. Building the
    # labextension creates unpredictable file names, and `data_file` wants to
    # explicitly list all files one by one.

    if needs_npm_build():
        print("==== rebuilding the JS code with npm, this might take a while ...")
        run_npm_build()

    setup(
        cmdclass={
            "bdist_egg": bdist_egg if "bdist_egg" in sys.argv else bdist_egg_disabled,
        },
        package_data={
            "chemiscope": [
                "sphinx/static/*",
                "streamlit/**/*",
            ]
        },
        data_files=[
            # this is what `jupyter nbextension install --sys-prefix` does
            (
                "share/jupyter/nbextensions/chemiscope",
                [
                    "python/jupyter/extension.js",
                    "python/jupyter/nbextension/chemiscope.min.js",
                ],
            ),
            # this is what `jupyter nbextension enable --sys-prefix` does
            (
                "etc/jupyter/nbconfig/notebook.d",
                [
                    "python/jupyter/chemiscope.json",
                ],
            ),
            # install the labextension
            (
                "share/jupyter/labextensions/chemiscope",
                list(
                    filter(
                        lambda f: os.path.isfile(f),
                        glob.glob("python/jupyter/labextension/*"),
                    )
                ),
            ),
            (
                "share/jupyter/labextensions/chemiscope/static",
                glob.glob("python/jupyter/labextension/static/*"),
            ),
        ],
    )
